# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Module with click utils of the aea cli."""
import os
from collections import OrderedDict
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Generator, List, Optional, Sequence, Type, Union, cast

import click
from click import argument, option

from aea.cli.registry.settings import (
    REGISTRY_LOCAL,
    REGISTRY_MIXED,
    REGISTRY_REMOTE,
    REMOTE_HTTP,
    REMOTE_IPFS,
)
from aea.cli.utils.config import get_or_create_cli_config, try_to_load_agent_config
from aea.cli.utils.constants import DUMMY_PACKAGE_ID
from aea.configurations.constants import (
    CONFIG_FILE_TO_PACKAGE_TYPE,
    CONNECTION,
    CONTRACT,
    DEFAULT_AEA_CONFIG_FILE,
    PROTOCOL,
    SKILL,
)
from aea.configurations.data_types import Dependency, PackageType, PublicId
from aea.crypto.registries import crypto_registry
from aea.helpers.io import open_file
from aea.package_manager.base import PACKAGE_SOURCE_RE


class ConnectionsOption(click.Option):
    """Click option for the --connections option in 'aea run'."""

    def type_cast_value(
        self, ctx: click.Context, value: str
    ) -> Optional[List[PublicId]]:
        """
        Parse the list of string passed through command line.

        E.g. from 'stub,local' to ['stub', 'local'].

        :param ctx: the click context
        :param value: the list of connection names, as a string.
        :return: list of public ids
        """
        if value is None:
            return None
        try:

            def arg_strip(s: str) -> str:
                return s.strip(" '\"")

            input_connection_ids = [
                arg_strip(s) for s in value.split(",") if arg_strip(s) != ""
            ]

            # remove duplicates, while preserving the order
            result = OrderedDict()  # type: OrderedDict[PublicId, None]
            for connection_id_string in input_connection_ids:
                connection_public_id = PublicId.from_str(connection_id_string)
                result[connection_public_id] = None
            return list(result.keys())
        except Exception:  # pragma: no cover
            raise click.BadParameter(value)


class PytestArgs(click.Option):
    """Custom Click option for parsing Pytest arguments."""

    def type_cast_value(
        self, ctx: click.Context, value: Optional[str]
    ) -> Sequence[str]:
        """Cast a string value to a sequence of Pytest arguments."""
        try:
            if value is None:
                return []
            return value.split(" ")
        except Exception:
            error_message = f"cannot split '{value}' into pytest arguments"
            raise click.BadParameter(error_message)


class PublicIdOrPathParameter(click.ParamType):
    """Define a public id parameter for Click applications."""

    def __init__(  # pylint: disable=useless-super-delegation
        self, *args: Any, **kwargs: Any
    ) -> None:
        """
        Initialize the Public Id parameter.

        Just forwards arguments to parent constructor.

        :param args: positional arguments
        :param kwargs: keyword arguments
        """
        super().__init__(*args, **kwargs)  # type: ignore

    def get_metavar(self, param: Any) -> str:
        """Return the metavar default for this param if it provides one."""
        return "PUBLIC_ID_OR_PATH"

    @staticmethod
    def _parse_public_id(value: str) -> Optional[PublicId]:
        """Parse public id from string."""

        try:
            return PublicId.from_str(value)
        except ValueError:
            return None

    @staticmethod
    def _parse_path(value: str) -> Optional[Path]:
        """Parse path from string."""
        path = Path(value).absolute()
        if path.is_dir():
            return path
        return None

    def convert(
        self, value: str, param: Any, ctx: Optional[click.Context]
    ) -> Union[PublicId, Path]:
        """Convert the value. This is not invoked for values that are `None` (the missing value)."""
        parsed_value: Optional[Union[PublicId, Path]]
        parsed_value = self._parse_public_id(value)
        if parsed_value is not None:
            return parsed_value

        parsed_value = self._parse_path(value)
        if parsed_value is None:
            self.fail(value, param, ctx)

        return cast(Path, parsed_value)


class PublicIdParameter(click.ParamType):
    """Define a public id parameter for Click applications."""

    def get_metavar(self, param: Any) -> str:
        """Return the metavar default for this param if it provides one."""
        return "PUBLIC_ID_OR_HASH"

    @staticmethod
    def _parse_public_id(value: str) -> Optional[PublicId]:
        """Parse extended public from string."""
        try:
            return PublicId.from_str(value)
        except ValueError:
            return None

    @staticmethod
    def _parse_hash(value: str) -> Optional[PublicId]:
        """Parse extended public from string."""
        try:
            return PublicId.from_json({**DUMMY_PACKAGE_ID.json, "package_hash": value})
        except ValueError:
            return None

    def convert(self, value: str, param: Any, ctx: Optional[click.Context]) -> PublicId:
        """Convert the value. This is not invoked for values that are `None` (the missing value)."""

        parsed = self._parse_public_id(value)
        if parsed is not None:
            return parsed

        parsed = self._parse_hash(value)
        if parsed is None:
            self.fail(value, param, ctx)

        return cast(PublicId, parsed)


class AgentDirectory(click.Path):
    """A click.Path, but with further checks  applications."""

    def __init__(self) -> None:
        """Initialize the agent directory parameter."""
        super().__init__(
            exists=True, file_okay=False, dir_okay=True, readable=True, writable=False
        )

    def get_metavar(self, param: Any) -> str:
        """Return the metavar default for this param if it provides one."""
        return "AGENT_DIRECTORY"  # pragma: no cover

    def convert(self, value: str, param: Any, ctx: click.Context) -> str:  # type: ignore
        """Convert the value. This is not invoked for values that are `None` (the missing value)."""
        cwd = os.getcwd()
        path = Path(value)
        try:
            # check that the target folder is an AEA project.
            os.chdir(path)
            with open_file(DEFAULT_AEA_CONFIG_FILE, mode="r", encoding="utf-8") as fp:
                ctx.obj.agent_config = ctx.obj.agent_loader.load(fp)
            try_to_load_agent_config(ctx.obj)
            # everything ok - return the parameter to the command
            return value
        except Exception:
            raise click.ClickException(
                "The name provided is not a path to an AEA project."
            )
        finally:
            os.chdir(cwd)


class PackagesSource(click.ParamType):
    """Click paramater for package sources."""

    def __init__(self) -> None:
        """Initialize the package source parameter."""

    def get_metavar(self, param: Any) -> str:
        """Return the metavar default for this param if it provides one."""
        return "SOURCE"  # pragma: no cover

    def convert(self, value: str, param: Any, ctx: click.Context) -> str:
        """Convert the value."""
        if PACKAGE_SOURCE_RE.match(value) is None:
            raise click.ClickException(
                f"Bad value provided for package source `{value}`"
            )
        return value


class PyPiDependency(click.ParamType):
    """Click parameter for PyPy dependency string"""

    def get_metavar(self, param: Any) -> str:
        """Return the metavar default for this param if it provides one."""
        return "DEPENDENCY"  # pragma: no cover

    def convert(self, value: str, param: Any, ctx: click.Context) -> Dependency:
        """Convert the value."""
        try:
            return Dependency.from_string(value)
        except ValueError as e:
            raise click.ClickException(str(e)) from e


class LedgerChoice(click.ParamType):
    """Ledger choice flag."""

    ALL = "all"

    def get_metavar(self, param: click.Parameter) -> str:
        """Get metavar name."""
        return "LEDGER_ID"

    def convert(
        self,
        value: str,
        param: Optional[click.Parameter] = None,
        ctx: Optional[click.Context] = None,
    ) -> str:
        """Convert to ledger id."""
        if len(crypto_registry.supported_ids) == 0:
            raise click.ClickException("No ledger installation found")

        if value not in crypto_registry.supported_ids and value != self.ALL:
            raise click.ClickException(
                f"Invalid identifier provided `{value}`; "
                f"Available ledger identifiers {crypto_registry.supported_ids}"
            )

        return value


def registry_flag(
    mark_default: bool = True,
    default_registry: Optional[str] = None,
) -> Callable:
    """Choice of one flag between: '--local/--remote/--mixed'."""

    default_registry = default_registry or (
        get_or_create_cli_config().get("registry_config", {}).get("default")
    )
    default_registry = default_registry or REGISTRY_LOCAL

    def wrapper(f: Callable) -> Callable:
        f = option(
            "--local",
            "registry",
            flag_value=REGISTRY_LOCAL,
            help="To use a local registry.",
            default=(REGISTRY_LOCAL == default_registry) and mark_default,
        )(f)
        f = option(
            "--remote",
            "registry",
            flag_value=REGISTRY_REMOTE,
            help="To use a remote registry.",
            default=(REGISTRY_REMOTE == default_registry) and mark_default,
        )(f)
        f = option(
            "--mixed",
            "registry",
            flag_value=REGISTRY_MIXED,
            help="To use a local and remote registries.",
            default=(REGISTRY_MIXED == default_registry) and mark_default,
        )(f)
        return f

    return wrapper


def remote_registry_flag(
    mark_default: bool = True,
    default_registry: Optional[str] = None,
) -> Callable:
    """Choice of one flag between: '--ipfs/--http'."""

    default_registry = default_registry or (
        get_or_create_cli_config()
        .get("registry_config", {})
        .get("settings", {})
        .get("remote", {})
        .get("default")
    )

    default_registry = default_registry or REMOTE_IPFS

    def wrapper(f: Callable) -> Callable:
        f = option(
            "--ipfs",
            "remote_registry",
            flag_value=REMOTE_IPFS,
            help="To use an IPFS registry.",
            default=(REMOTE_IPFS == default_registry) and mark_default,
        )(f)
        f = option(
            "--http",
            "remote_registry",
            flag_value=REMOTE_HTTP,
            help="To use an HTTP registry.",
            default=(REMOTE_HTTP == default_registry) and mark_default,
        )(f)
        return f

    return wrapper


def registry_path_option(f: Callable) -> Callable:
    """Add registry path aea option."""
    return option(
        "--registry-path",
        type=click.Path(dir_okay=True, exists=True, file_okay=False),
        required=False,
        help="Provide a local registry directory full path.",
    )(f)


def component_flag(
    wrap_public_id: bool = False,
) -> Callable:
    """Wraps a command with component types argument"""

    def wrapper(f: Callable) -> Callable:
        if wrap_public_id:
            f = argument("public_id", type=PublicIdParameter(), required=True)(f)

        f = argument(
            "component_type",
            type=click.Choice((CONNECTION, CONTRACT, PROTOCOL, SKILL)),
            required=True,
        )(f)

        return f

    return wrapper


def password_option(confirmation_prompt: bool = False, **kwargs) -> Callable:  # type: ignore
    """Decorator to ask for input if -p flag was provided or use --password to set password value in command line."""

    def callback(ctx, _, value: bool) -> bool:  # type: ignore
        if value is True:
            ctx.params["password"] = ctx.params.get("password") or click.prompt(
                "Enter password",
                hide_input=True,
                confirmation_prompt=confirmation_prompt,
            )
        return value

    def wrap(fn):  # type: ignore
        return click.option(
            "-p",
            is_flag=True,
            type=bool,
            callback=callback,
            expose_value=False,
            help="Ask for password interactively",
        )(
            click.option(
                "--password",
                type=str,
                is_eager=True,
                metavar="PASSWORD",
                help="Set password for key encryption/decryption",
                **kwargs,
            )(fn)
        )  # type: ignore

    return wrap


def determine_package_type_for_directory(package_dir: Path) -> PackageType:
    """
    Find package type for the package directory by checking config file names.

    :param package_dir: package dir to determine package type:

    :return: PackageType
    """
    config_files = list(
        set(os.listdir(str(package_dir))).intersection(
            set(CONFIG_FILE_TO_PACKAGE_TYPE.keys())
        )
    )

    if len(config_files) > 1:
        raise ValueError(
            f"Too many config files in the directory, only one has to present!: {', '.join(config_files)}"
        )
    if len(config_files) == 0:
        raise ValueError(
            f"No package config file found in `{str(package_dir)}`. Incorrect directory?"
        )

    config_file = config_files[0]
    package_type = PackageType(CONFIG_FILE_TO_PACKAGE_TYPE[config_file])

    return package_type


@contextmanager
def reraise_as_click_exception(*exceptions: Type[Exception]) -> Generator:
    """Reraise exceptions as ClickException"""
    try:
        yield
    except exceptions as e:
        raise click.ClickException(str(e))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains a test for aea.test_tools."""

import os
from pathlib import Path
from typing import cast
from unittest.mock import patch

import click
import pytest
from _pytest.capture import CaptureFixture  # type: ignore

import aea
from aea.cli.core import cli
from aea.test_tools.click_testing import CliRunner, CliTest
from aea.test_tools.utils import copy_class


def test_invoke():
    """Test runner invoke method."""
    cli_runner = CliRunner()

    result = cli_runner.invoke(cli, ["--help"])
    assert (
        "Command-line tool for setting up an Autonomous Economic Agent" in result.output
    )

    result = cli_runner.invoke(cli, "--help")
    assert (
        "Command-line tool for setting up an Autonomous Economic Agent" in result.output
    )


def test_invoke_error():
    """Test runner invoke method raises an error."""
    cli_runner = CliRunner()

    with patch.object(cli, "main", side_effect=SystemExit(1)):
        result = cli_runner.invoke(cli, ["--help"])
        assert result.exit_code == 1

    with patch.object(cli, "main", side_effect=SystemExit(object())):
        result = cli_runner.invoke(cli, ["--help"])
        assert result.exit_code == 1


def test_catch_exception():
    """Test runner invoke method raises an exception and its propogated."""
    cli_runner = CliRunner()

    # True
    with patch.object(cli, "main", side_effect=ValueError("expected")):
        result = cli_runner.invoke(cli, ["--help"])
        assert result.exit_code == 1

    # False
    with pytest.raises(ValueError, match="expected"):
        with patch.object(cli, "main", side_effect=ValueError("expected")):
            cli_runner.invoke(cli, ["--help"], catch_exceptions=False)


def test_mix_std_err_False():
    """Test stderr and stdout not mixed."""
    cli_runner = CliRunner(mix_stderr=False)

    result = cli_runner.invoke(cli, "-v DEBUG run")
    assert result.exit_code == 1
    # check for access, no exception should be raised
    assert result.stderr is not None


def test_mix_std_err_True():
    """Test stderr and stdout are mixed."""
    cli_runner = CliRunner(mix_stderr=True)

    result = cli_runner.invoke(cli, "-v DEBUG run")
    assert result.exit_code == 1

    with pytest.raises(ValueError, match="stderr not separately captured"):
        assert result.stderr


def test_click_version():
    """Test click version"""

    message = """
    When this tests fails you need to ensure that the current versions implementation
    of the click.testing.CliRunner remains compatible with our monkey-patched version
    """
    assert click.__version__ == "8.0.2", message


@pytest.mark.parametrize("mix_stderr", [True, False])
def test_capfd_on_cli_runner(mix_stderr: bool, capfd: CaptureFixture):
    """Test setting capfd on CliRunner to redirect streams"""

    def run_cli_command_and_assert() -> None:
        result = cli_runner.invoke(cli, ["--help"], standalone_mode=False)
        expected = "Command-line tool for setting up an Autonomous Economic Agent"
        assert expected in result.stdout
        if mix_stderr:
            with pytest.raises(ValueError, match="stderr not separately captured"):
                assert result.stderr

    cli_runner = CliRunner(mix_stderr=mix_stderr)

    with patch.object(capfd, "readouterr", wraps=capfd.readouterr) as m:
        # streams captured via CliRunner.isolation context manager
        run_cli_command_and_assert()
        m.assert_not_called()

        # streams captured via pytest capfd fixture
        cli_runner.capfd = capfd
        run_cli_command_and_assert()
        m.assert_called_once()


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(input="input"),
        dict(env={"key": "value"}),
        dict(color=True),
    ],
)
def test_cli_runner_invoke_raises(kwargs, capfd: CaptureFixture):
    """Test CliRunner with capfd raises"""

    cli_runner = CliRunner()
    cli_runner.capfd = capfd
    with pytest.raises(
        NotImplementedError,
        match="Cannot use capfd in conjunction with `input`, `env` or `color`.",
    ):
        cli_runner.invoke(cli, ["--help"], standalone_mode=False, **kwargs)


class TestCliTest:
    """Test CliTest"""

    def setup(self) -> None:
        """Setup test"""
        # `copy` the class to avoid test interference
        self.test_cls = cast(CliTest, copy_class(CliTest))

    def setup_test(self) -> CliTest:
        """Setup test"""

        self.test_cls.setup_class()
        test_instance = self.test_cls()  # type: ignore
        test_instance.setup()
        return test_instance

    def test_setup_cls_and_setup(self) -> None:
        """Test setup_class and setup"""

        self.test_cls.setup_class()
        assert isinstance(self.test_cls._CliTest__cli_runner, CliRunner)  # type: ignore
        assert self.test_cls._CliTest__cli.name == "aea"  # type: ignore
        assert not hasattr(self.test_cls, "_t")

        test_instance = self.test_cls()  # type: ignore
        test_instance.setup()
        assert isinstance(test_instance._t, Path)

    def test_teardown_and_teardown_cls(self) -> None:
        """Test teardown and teardown_class"""

        test_instance = self.setup_test()
        cwd = Path.cwd()

        assert not test_instance._t == cwd
        os.chdir(test_instance._t)
        assert test_instance._t == Path.cwd()
        test_instance.teardown()
        assert not hasattr(self.test_cls, "_t")

        test_instance.teardown_class()
        assert Path.cwd() == cwd

    def test_run_cli(self) -> None:
        """Test run_cli"""

        test_instance = self.setup_test()
        result = test_instance.run_cli("--version")
        assert result.exit_code == 0
        assert f"aea, version {aea.__version__}" in result.stdout

    def test_run_cli_subprocess(self) -> None:
        """Test run_cli_subprocess"""

        test_instance = self.setup_test()
        result = test_instance.run_cli_subprocess("--version")
        assert result.exit_code == 0
        assert f"aea, version {aea.__version__}" in result.stdout

    def test_run_cli_failure(self) -> None:
        """Test run_cli failure"""

        test_instance = self.setup_test()
        result = test_instance.run_cli("non-existent-command")
        assert result.exit_code == 2
        assert "No such command 'non-existent-command'" in result.stdout

    def test_run_cli_subprocess_failure(self) -> None:
        """Test run_cli_subprocess failure"""

        test_instance = self.setup_test()
        result = test_instance.run_cli_subprocess("non-existent-command")
        assert result.exit_code == 2
        assert "No such command 'non-existent-command'" in result.output



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains the behaviour of a erc1155 deploy skill AEA."""

from typing import Any, Optional, cast

from aea.helpers.search.models import Description
from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.erc1155_deploy.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    LedgerApiDialogues,
    OefSearchDialogues,
)
from packages.fetchai.skills.erc1155_deploy.strategy import Strategy
from packages.valory.connections.ledger.connection import (
    PUBLIC_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage


DEFAULT_SERVICES_INTERVAL = 30.0
DEFAULT_MAX_SOEF_REGISTRATION_RETRIES = 5
LEDGER_API_ADDRESS = str(LEDGER_CONNECTION_PUBLIC_ID)


class ServiceRegistrationBehaviour(TickerBehaviour):
    """This class implements a behaviour."""

    def __init__(self, **kwargs: Any) -> None:
        """Initialise the behaviour."""
        services_interval = kwargs.pop(
            "services_interval", DEFAULT_SERVICES_INTERVAL
        )  # type: int
        self._max_soef_registration_retries = kwargs.pop(
            "max_soef_registration_retries", DEFAULT_MAX_SOEF_REGISTRATION_RETRIES
        )  # type: int
        super().__init__(tick_interval=services_interval, **kwargs)
        self.is_registered = False
        self.registration_in_progress = False
        self.failed_registration_msg = None  # type: Optional[OefSearchMessage]
        self._nb_retries = 0

    def setup(self) -> None:
        """Implement the setup."""
        self._request_balance()
        strategy = cast(Strategy, self.context.strategy)
        if not strategy.is_contract_deployed:
            self._request_contract_deploy_transaction()

    def act(self) -> None:
        """Implement the act."""
        self._retry_failed_registration()

        strategy = cast(Strategy, self.context.strategy)
        if not strategy.is_behaviour_active:
            return

        if strategy.is_contract_deployed and not strategy.is_tokens_created:
            self._request_token_create_transaction()
        elif (
            strategy.is_contract_deployed
            and strategy.is_tokens_created
            and not strategy.is_tokens_minted
        ):
            self._request_token_mint_transaction()
        elif (
            strategy.is_contract_deployed
            and strategy.is_tokens_created
            and strategy.is_tokens_minted
            and not self.registration_in_progress
            and not self.is_registered
        ):
            self.registration_in_progress = True
            self._register_agent()

    def teardown(self) -> None:
        """Implement the task teardown."""
        self._unregister_service()
        self._unregister_agent()

    def _retry_failed_registration(self) -> None:
        """Retry a failed registration."""
        if self.failed_registration_msg is not None:
            self._nb_retries += 1
            if self._nb_retries > self._max_soef_registration_retries:
                self.context.is_active = False
                return

            oef_search_dialogues = cast(
                OefSearchDialogues, self.context.oef_search_dialogues
            )
            oef_search_msg, _ = oef_search_dialogues.create(
                counterparty=self.failed_registration_msg.to,
                performative=self.failed_registration_msg.performative,
                service_description=self.failed_registration_msg.service_description,
            )
            self.context.outbox.put_message(message=oef_search_msg)
            self.context.logger.info(
                f"Retrying registration on SOEF. Retry {self._nb_retries} out of {self._max_soef_registration_retries}."
            )

            self.failed_registration_msg = None

    def _request_balance(self) -> None:
        """Request ledger balance."""
        strategy = cast(Strategy, self.context.strategy)
        ledger_api_dialogues = cast(
            LedgerApiDialogues, self.context.ledger_api_dialogues
        )
        ledger_api_msg, _ = ledger_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_BALANCE,
            ledger_id=strategy.ledger_id,
            address=cast(str, self.context.agent_addresses.get(strategy.ledger_id)),
        )
        self.context.outbox.put_message(message=ledger_api_msg)

    def _request_contract_deploy_transaction(self) -> None:
        """Request contract deploy transaction"""
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=strategy.contract_id,
            callable="get_deploy_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {"deployer_address": self.context.agent_address, "gas": strategy.gas}
            ),
        )
        contract_api_dialogue = cast(
            ContractApiDialogue,
            contract_api_dialogue,
        )
        contract_api_dialogue.terms = strategy.get_deploy_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting contract deployment transaction...")

    def _request_token_create_transaction(self) -> None:
        """Request token create transaction."""
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=strategy.contract_id,
            contract_address=strategy.contract_address,
            callable="get_create_batch_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.context.agent_address,
                    "token_ids": strategy.token_ids,
                    "gas": strategy.gas,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue)
        contract_api_dialogue.terms = strategy.get_create_token_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting create batch transaction...")

    def _request_token_mint_transaction(self) -> None:
        """Request token mint transaction."""
        strategy = cast(Strategy, self.context.strategy)
        strategy.is_behaviour_active = False
        contract_api_dialogues = cast(
            ContractApiDialogues, self.context.contract_api_dialogues
        )
        contract_api_msg, contract_api_dialogue = contract_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            ledger_id=strategy.ledger_id,
            contract_id=strategy.contract_id,
            contract_address=strategy.contract_address,
            callable="get_mint_batch_transaction",
            kwargs=ContractApiMessage.Kwargs(
                {
                    "deployer_address": self.context.agent_address,
                    "recipient_address": self.context.agent_address,
                    "token_ids": strategy.token_ids,
                    "mint_quantities": strategy.mint_quantities,
                    "gas": strategy.gas,
                }
            ),
        )
        contract_api_dialogue = cast(ContractApiDialogue, contract_api_dialogue)
        contract_api_dialogue.terms = strategy.get_mint_token_terms()
        self.context.outbox.put_message(message=contract_api_msg)
        self.context.logger.info("requesting mint batch transaction...")

    def _register(self, description: Description, logger_msg: str) -> None:
        """
        Register something on the SOEF.

        :param description: the description of what is being registered
        :param logger_msg: the logger message to print after the registration
        """
        oef_search_dialogues = cast(
            OefSearchDialogues, self.context.oef_search_dialogues
        )
        oef_search_msg, _ = oef_search_dialogues.create(
            counterparty=self.context.search_service_address,
            performative=OefSearchMessage.Performative.REGISTER_SERVICE,
            service_description=description,
        )
        self.context.outbox.put_message(message=oef_search_msg)
        self.context.logger.info(logger_msg)

    def _register_agent(self) -> None:
        """Register the agent's location."""
        strategy = cast(Strategy, self.context.strategy)
        description = strategy.get_location_description()
        self._register(description, "registering agent on SOEF.")

    def register_service(self) -> None:
        """Register the agent's service."""
        strategy = cast(Strategy, self.context.strategy)
        description = strategy.get_register_service_description()
        self._register(description, "registering agent's service on the SOEF.")

    def register_genus(self) -> None:
        """Register the agent's personality genus."""
        strategy = cast(Strategy, self.context.strategy)
        description = strategy.get_register_personality_description()
        self._register(
            description, "registering agent's personality genus on the SOEF."
        )

    def register_classification(self) -> None:
        """Register the agent's personality classification."""
        strategy = cast(Strategy, self.context.strategy)
        description = strategy.get_register_classification_description()
        self._register(
            description, "registering agent's personality classification on the SOEF."
        )

    def _unregister_service(self) -> None:
        """Unregister service from the SOEF."""
        strategy = cast(Strategy, self.context.strategy)
        description = strategy.get_unregister_service_description()
        oef_search_dialogues = cast(
            OefSearchDialogues, self.context.oef_search_dialogues
        )
        oef_search_msg, _ = oef_search_dialogues.create(
            counterparty=self.context.search_service_address,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            service_description=description,
        )
        self.context.outbox.put_message(message=oef_search_msg)
        self.context.logger.info("unregistering service from SOEF.")

    def _unregister_agent(self) -> None:
        """Unregister agent from the SOEF."""
        strategy = cast(Strategy, self.context.strategy)
        description = strategy.get_location_description()
        oef_search_dialogues = cast(
            OefSearchDialogues, self.context.oef_search_dialogues
        )
        oef_search_msg, _ = oef_search_dialogues.create(
            counterparty=self.context.search_service_address,
            performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
            service_description=description,
        )
        self.context.outbox.put_message(message=oef_search_msg)
        self.context.logger.info("unregistering agent from SOEF.")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""ACN Docker Image."""

import logging
import socket
import time
from typing import Dict, List

from aea.exceptions import enforce
from aea.test_tools.docker_image import Container, DockerClient, DockerImage


META_ADDRESS = "0.0.0.0"  # nosec

# created agent: bootstrap_peer
#     private key: 7f669ab5eee5719e385f7aeb1973769fc75b7cbbe0850ca16c4eabe84e01afbd
#     public key:  0270475f9b78c0285a6ac6067582f5e159ec147ccb03aee16a32731f68920b1ae8
#     PeerID:      16Uiu2HAm2yxmLQTZTrxjo5c4k5ka8AVMcpeD5zMMeasE6xDw1YQw
#
# created agent: entry_node_1
#     private key: 5b8d3be9f27489040a01adc2b746b9a5f9d32ed843d99cf7d2995c8140636190
#     public key:  02197b55d736bd242311aaabb485f9db40881349873bb13e8b60c8a130ecb341d8
#     PeerID:      16Uiu2HAkw99FW2GKb2qs24eLgfXSSUjke1teDaV9km63Fv3UGdnF
#
# created agent: entry_node_2
#     private key: cc096b7be575c11d3d3d2f8a9c9be9bd59b351317b2a114ef7e014cc5a92508e
#     public key:  0287ee61e8f939aeaa69bd7156463d698f8e74a3e1d5dd20cce997970f13ad4f12
#     PeerID:      16Uiu2HAm4aHr1iKR323tca8Zu8hKStEEVwGkE2gtCJw49S3gbuVj


GENESIS_MADDR = f"/dns4/{META_ADDRESS}/tcp/9000/p2p/16Uiu2HAm2yxmLQTZTrxjo5c4k5ka8AVMcpeD5zMMeasE6xDw1YQw"

BOOTSTRAP: Dict[str, str] = dict(
    AEA_P2P_ID="7f669ab5eee5719e385f7aeb1973769fc75b7cbbe0850ca16c4eabe84e01afbd",
    AEA_P2P_URI_PUBLIC=f"{META_ADDRESS}:9000",
    AEA_P2P_URI=f"{META_ADDRESS}:9000",
    AEA_P2P_DELEGATE_URI=f"{META_ADDRESS}:11000",
    AEA_P2P_URI_MONITORING=f"{META_ADDRESS}:8080",
    ACN_LOG_FILE="/acn/libp2p_node.log",
)

NODE1: Dict[str, str] = dict(
    AEA_P2P_ID="5b8d3be9f27489040a01adc2b746b9a5f9d32ed843d99cf7d2995c8140636190",
    AEA_P2P_URI_PUBLIC=f"{META_ADDRESS}:9001",
    AEA_P2P_URI=f"{META_ADDRESS}:9001",
    AEA_P2P_DELEGATE_URI=f"{META_ADDRESS}:11001",
    AEA_P2P_URI_MONITORING=f"{META_ADDRESS}:8081",
    AEA_P2P_ENTRY_URIS=GENESIS_MADDR,
    ACN_LOG_FILE="/acn/libp2p_node.log",
)

NODE2: Dict[str, str] = dict(
    AEA_P2P_ID="cc096b7be575c11d3d3d2f8a9c9be9bd59b351317b2a114ef7e014cc5a92508e",
    AEA_P2P_URI_PUBLIC=f"{META_ADDRESS}:9002",
    AEA_P2P_URI=f"{META_ADDRESS}:9002",
    AEA_P2P_DELEGATE_URI=f"{META_ADDRESS}:11002",
    AEA_P2P_URI_MONITORING=f"{META_ADDRESS}:8082",
    AEA_P2P_ENTRY_URIS=GENESIS_MADDR,
    ACN_LOG_FILE="/acn/libp2p_node.log",
)


class ACNNodeDockerImage(DockerImage):
    """Wrapper to ACNNode Docker image."""

    uris: List = [
        "AEA_P2P_URI_PUBLIC",
        "AEA_P2P_URI",
        "AEA_P2P_DELEGATE_URI",
        "AEA_P2P_URI_MONITORING",
    ]

    nodes = ["bootstrap"]

    def __init__(
        self,
        client: DockerClient,
        config: Dict,
    ):
        """
        Initialize the ACNNode Docker image.

        :param client: the Docker client.
        :param config: optional configuration to command line.
        """
        super().__init__(client)
        self._config = config
        self._extra_hosts = {name: "host-gateway" for name in self.nodes}

    @property
    def tag(self) -> str:
        """Get the image tag."""
        return "valory/open-acn-node:latest"

    @property
    def ports(self) -> List[str]:
        """Ports"""
        return [self._config[uri].split(":")[1] for uri in self.uris]

    def _make_ports(self) -> Dict:
        """Make ports dictionary for Docker."""

        return {f"{p}/tcp": (META_ADDRESS, p) for p in self.ports}

    def create(self) -> Container:
        """Create the container."""
        container = self._client.containers.run(
            image=self.tag,
            command=["--config-from-env"],
            detach=True,
            ports=self._make_ports(),
            environment=self._config,
            extra_hosts=self._extra_hosts,
        )
        return container

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait until the image is up."""

        i, to_be_connected = 0, {self._config[uri] for uri in self.uris}
        while i < max_attempts and to_be_connected:
            i += 1
            for uri in to_be_connected:
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    host, port = uri.split(":")
                    result = sock.connect_ex((host, int(port)))
                    sock.close()
                    enforce(result == 0, "")
                    to_be_connected.remove(uri)
                    logging.info(f"URI ready: {uri}")
                    break
                except Exception:  # pylint: disable=broad-except
                    logging.error(
                        f"Attempt {i} failed on {uri}. Retrying in {sleep_rate} seconds..."
                    )
                    time.sleep(sleep_rate)

        return not to_be_connected


class ACNWithBootstrappedEntryNodesDockerImage(ACNNodeDockerImage):  # noqa: F401
    """ACN with bootstrapped entry nodes"""

    nodes = ["bootstrap", "entry_node_1", "entry_node_2"]
    configs = [BOOTSTRAP, NODE1, NODE2]

    def create(self) -> List[Container]:
        """Instantiate the image in many containers, parametrized."""

        containers = []

        for name, config in zip(self.nodes, self.configs):
            kwargs = dict(
                image=self.tag,
                hostname=name,
                command=["--config-from-env"],
                detach=True,
                network="host",
                environment=config,
                extra_hosts=self._extra_hosts,
            )
            containers.append(self._client.containers.run(**kwargs))

        return containers

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """Wait - this is container specific (using self._config) so doesn't work"""
        time.sleep(sleep_rate)
        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea` sub-commands."""

import aea
from aea.cli import cli

from tests.conftest import CliRunner


def test_no_argument():
    """Test that if we run the cli tool without arguments, it exits gracefully."""
    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert result.exit_code == 0


def test_flag_version():
    """Test that the flag '--version' works correctly."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.stdout == "aea, version {}\n".format(aea.__version__)


def test_flag_help():
    """Test that the flag '--help' works correctly."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert (
        result.stdout
        == """Usage: aea [OPTIONS] COMMAND [ARGS]...

  Command-line tool for setting up an Autonomous Economic Agent (AEA).

Options:
  --version                     Show the version and exit.
  -v, --verbosity LVL           One of NOTSET, DEBUG, INFO, WARNING, ERROR,
                                CRITICAL, OFF
  -s, --skip-consistency-check  Skip consistency checks of agent during command
                                execution.
  --registry-path DIRECTORY     Provide a local registry directory full path.
  --help                        Show this message and exit.

Commands:
  add                     Add a package to the agent.
  add-key                 Add a private key to the wallet of the agent.
  build                   Build the agent and its components.
  check-packages          Run different checks on AEA packages.
  config                  Read or modify a configuration of the agent.
  create                  Create a new agent.
  delete                  Delete an agent.
  eject                   Eject a vendor package of the agent.
  fetch                   Fetch an agent from the registry.
  fingerprint             Fingerprint a non-vendor package of the agent.
  freeze                  Get the dependencies of the agent.
  generate                Generate a package for the agent.
  generate-all-protocols  Generate all protocols.
  generate-key            Generate a private key and place it in a file.
  generate-wealth         Generate wealth for the agent on a test network.
  get-address             Get the address associated with a private key of...
  get-multiaddress        Get the multiaddress associated with a private...
  get-public-key          Get the public key associated with a private key...
  get-wealth              Get the wealth associated with the private key of...
  hash                    Hashing utils.
  init                    Initialize your AEA configurations.
  install                 Install the dependencies of the agent.
  ipfs                    IPFS Commands
  issue-certificates      Issue certificates for connections that require...
  launch                  Launch many agents at the same time.
  list                    List the installed packages of the agent.
  local-registry-sync     Sync packages between packages.json and a local...
  login                   Login to the registry account.
  logout                  Logout from the registry account.
  packages                Local package manager.
  publish                 Publish the agent to the registry.
  push                    Push a non-vendor package of the agent to the...
  push-all                Push all available packages to a registry.
  register                Create a new registry account.
  remove                  Remove a package from the agent.
  remove-key              Remove a private key from the wallet of the agent.
  reset_password          Reset the password of the registry account.
  run                     Run the agent.
  scaffold                Scaffold a package for the agent.
  search                  Search for packages in the registry.
  test                    Run tests of an AEA project.
  transfer                Transfer wealth associated with a private key of...
  upgrade                 Upgrade the packages of the agent.
"""
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for CLI Registry fetch methods."""
import os
import shutil
import sys
from abc import ABC
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase, mock

import click
import pytest
from aea_cli_ipfs.ipfs_utils import DownloadError, IPFSTool
from click import ClickException

from aea.cli import cli
from aea.cli.fetch import _is_version_correct, fetch_agent_locally
from aea.cli.registry.settings import REMOTE_HTTP, REMOTE_IPFS
from aea.cli.utils.context import Context
from aea.configurations.base import PublicId
from aea.helpers.base import cd
from aea.test_tools.test_cases import (
    AEATestCaseMany,
    AEATestCaseManyFlaky,
    BaseAEATestCase,
)

from tests.conftest import (
    CLI_LOG_OPTION,
    CliRunner,
    MAX_FLAKY_RERUNS,
    MY_FIRST_AEA_PUBLIC_ID,
    PACKAGES_DIR,
    TEST_IPFS_REGISTRY_CONFIG,
)
from tests.test_cli.tools_for_testing import ContextMock, PublicIdMock


def _raise_click_exception(*args, **kwargs):
    raise ClickException("Message")


@mock.patch("builtins.open", mock.mock_open())
@mock.patch("aea.cli.utils.decorators._cast_ctx")
@mock.patch("aea.cli.fetch.os.path.join", return_value="joined-path")
@mock.patch("aea.cli.fetch.try_get_item_source_path", return_value="path")
@mock.patch("aea.cli.fetch.try_to_load_agent_config")
class FetchAgentLocallyTestCase(TestCase):
    """Test case for fetch_agent_locally method."""

    @mock.patch("aea.cli.fetch._is_version_correct", return_value=True)
    @mock.patch("aea.cli.fetch.os.path.exists", return_value=False)
    @mock.patch("aea.cli.fetch.copy_tree")
    def test_fetch_agent_locally_positive(self, copy_tree, *mocks):
        """Test for fetch_agent_locally method positive result."""
        ctx = ContextMock()
        ctx.config["is_local"] = True
        fetch_agent_locally(ctx, PublicIdMock(), alias="some-alias")
        copy_tree.assert_called_once_with("path", "joined-path")

    @mock.patch("aea.cli.fetch._is_version_correct", return_value=True)
    @mock.patch("aea.cli.fetch.os.path.exists", return_value=True)
    @mock.patch("aea.cli.fetch.copy_tree")
    def test_fetch_agent_locally_already_exists(self, *mocks):
        """Test for fetch_agent_locally method agent already exists."""
        ctx = ContextMock()
        ctx.config["is_local"] = True
        with self.assertRaises(ClickException):
            fetch_agent_locally(ctx, PublicIdMock())

    @mock.patch("aea.cli.fetch._is_version_correct", return_value=False)
    @mock.patch("aea.cli.fetch.os.path.exists", return_value=True)
    @mock.patch("aea.cli.fetch.copy_tree")
    def test_fetch_agent_locally_incorrect_version(self, *mocks):
        """Test for fetch_agent_locally method incorrect agent version."""
        ctx = ContextMock()
        ctx.config["is_local"] = True
        with self.assertRaises(ClickException):
            fetch_agent_locally(ctx, PublicIdMock())

    @mock.patch("aea.cli.fetch._is_version_correct", return_value=True)
    @mock.patch("aea.cli.fetch.add_item")
    @mock.patch("aea.cli.fetch.os.path.exists", return_value=False)
    @mock.patch("aea.cli.fetch.copy_tree")
    def test_fetch_agent_locally_with_deps_positive(self, *mocks):
        """Test for fetch_agent_locally method with deps positive result."""
        public_id = PublicIdMock.from_str("author/name:0.1.0")
        ctx_mock = ContextMock(
            connections=[public_id],
            protocols=[public_id],
            skills=[public_id],
            contracts=[public_id],
        )
        ctx_mock.config["is_local"] = True
        fetch_agent_locally(ctx_mock, PublicIdMock())

    @mock.patch("aea.cli.fetch._is_version_correct", return_value=True)
    @mock.patch("aea.cli.fetch.os.path.exists", return_value=False)
    @mock.patch("aea.cli.fetch.copy_tree")
    @mock.patch("aea.cli.fetch.add_item", _raise_click_exception)
    def test_fetch_agent_locally_with_deps_fail(self, *mocks):
        """Test for fetch_agent_locally method with deps ClickException catch."""
        public_id = PublicIdMock.from_str("author/name:0.1.0")
        ctx_mock = ContextMock(
            connections=[public_id],
            protocols=[public_id],
            skills=[public_id],
            contracts=[public_id],
        )
        ctx_mock.config["is_local"] = True
        with self.assertRaises(ClickException):
            fetch_agent_locally(ctx_mock, PublicIdMock())


@mock.patch("aea.cli.fetch.fetch_agent_remote")
@mock.patch("aea.cli.fetch.fetch_agent_locally")
class FetchCommandTestCase(TestCase):
    """Test case for CLI fetch command."""

    def setUp(self):
        """Set it up."""
        self.runner = CliRunner()

    def test_fetch_positive_mixed(self, *mocks):
        """Test for CLI push connection positive result."""
        self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fetch", "author/name:0.1.0"],
            standalone_mode=False,
        )

    def test_fetch_positive_local(self, *mocks):
        """Test for CLI push connection positive result."""
        self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fetch", "--local", "author/name:0.1.0"],
            standalone_mode=False,
        )

    def test_fetch_positive_remote(self, *mocks):
        """Test for CLI push connection positive result."""
        self.runner.invoke(
            cli,
            [*CLI_LOG_OPTION, "fetch", "--remote", "author/name:0.1.0"],
            standalone_mode=False,
        )


class IsVersionCorrectTestCase(TestCase):
    """Test case for _is_version_correct method."""

    def test__is_version_correct_positive(self):
        """Test for _is_version_correct method positive result."""
        public_id = PublicId("author", "package", "0.1.0")
        ctx_mock = ContextMock(version=public_id.version)
        ctx_mock.agent_config.public_id = public_id
        result = _is_version_correct(ctx_mock, public_id)
        self.assertTrue(result)

    def test__is_version_correct_negative(self):
        """Test for _is_version_correct method negative result."""
        public_id_a = PublicId("author", "package", "0.1.0")
        public_id_b = PublicId("author", "package", "0.1.1")
        ctx_mock = ContextMock(version=public_id_b.version)
        ctx_mock.agent_config.public_id = public_id_b
        result = _is_version_correct(ctx_mock, public_id_a)
        self.assertFalse(result)


@pytest.mark.skip(reason="https://agents-registry.prod.fetch-ai.com/ is down")
class TestFetchFromRemoteRegistryHTTP(AEATestCaseManyFlaky):
    """Test case for fetch agent command from Registry."""

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_fetch_agent_from_remote_registry_positive(self):
        """Test fetch agent from Registry for positive result."""

        with mock.patch(
            "aea.cli.utils.config.get_or_create_cli_config",
            return_value=TEST_IPFS_REGISTRY_CONFIG,
        ), mock.patch(
            "aea.cli.registry.utils.get_or_create_cli_config",
            return_value=TEST_IPFS_REGISTRY_CONFIG,
        ), mock.patch(
            "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_HTTP
        ):
            self.run_cli_command(
                "--skip-consistency-check",
                "fetch",
                "fetchai/my_first_aea:0.28.4",
                "--remote",
            )
        assert "my_first_aea" in os.listdir(self.t)


class TestFetchFromRemoteRegistryIPFS(AEATestCaseManyFlaky):
    """Test case for fetch agent command from Registry."""

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_fetch_agent_from_remote_registry_positive(self):
        """Test fetch agent from Registry for positive result."""

        with mock.patch(
            "aea.cli.utils.config.get_or_create_cli_config",
            return_value=TEST_IPFS_REGISTRY_CONFIG,
        ), mock.patch(
            "aea.cli.registry.utils.get_or_create_cli_config",
            return_value=TEST_IPFS_REGISTRY_CONFIG,
        ), mock.patch(
            "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_IPFS
        ):
            self.run_cli_command(
                "fetch",
                "open_aea/my_first_aea:0.1.0:bafybeiewms67jpwf46u4wwh6tbzedsi5jffajnywgydeo5nlvvr6pcz2zm",
                "--remote",
            )
        assert "my_first_aea" in os.listdir(self.t)


class TestFetchMixedModeFallsBackCorrectly(AEATestCaseManyFlaky):
    """Test fetch command when registry fetch fails falls back to remote fetch."""

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    @mock.patch(
        "aea.cli.fetch.fetch_agent_remote",
    )
    @mock.patch("aea.cli.fetch.fetch_agent_locally", side_effect=ClickException(""))
    def test_fetch_agent_from_local_registry_falls_back_to_remote(
        self, local_fetch, remote_fetch
    ):
        """Test fetch agent from Registry for positive result."""
        self.run_cli_command("fetch", "--mixed", str(MY_FIRST_AEA_PUBLIC_ID))
        remote_fetch.assert_called()


class TestFetchMixedModeLocalFirst(AEATestCaseManyFlaky):
    """Test fetch command when registry fetch runs local only."""

    @pytest.mark.integration
    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    @mock.patch("aea.cli.fetch.fetch_agent_remote")
    @mock.patch("aea.cli.fetch.fetch_agent_locally")
    def test_fetch_agent_from_local_registry_ok_remote_not_called(
        self, local_fetch: mock.Mock, remote_fetch: mock.Mock
    ):
        """Test fetch agent from Registry for positive result."""
        self.run_cli_command("fetch", "--mixed", str(MY_FIRST_AEA_PUBLIC_ID))
        local_fetch.assert_called()
        remote_fetch.assert_not_called()


class TestFetchLatestVersion(AEATestCaseMany):
    """Test case for fetch agent, latest version."""

    def test_fetch_agent_latest(self):
        """Test fetch agent, latest version."""
        self.run_cli_command(
            "fetch", "--local", str(MY_FIRST_AEA_PUBLIC_ID.to_latest())
        )
        assert "my_first_aea" in os.listdir(self.t)


class TestFetchAgentMixed(BaseAEATestCase):
    """Test 'aea fetch' in mixed mode."""

    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    @pytest.mark.integration
    @mock.patch("aea.cli.fetch.fetch_agent_remote")
    @mock.patch(
        "aea.cli.add.find_item_locally_or_distributed",
        side_effect=click.ClickException(""),
    )
    @mock.patch(
        "aea.cli.fetch.fetch_agent_locally",
        side_effect=click.ClickException(""),
    )
    def test_fetch_mixed(
        self, _mock_fetch_locally, _mock_fetch_agent_locally, mock_fetch_package
    ) -> None:
        """Test fetch in mixed mode."""
        self.run_cli_command(
            "-v", "DEBUG", "fetch", "--mixed", str(MY_FIRST_AEA_PUBLIC_ID.to_latest())
        )
        mock_fetch_package.assert_called()


class BaseTestFetchAgentError(BaseAEATestCase, ABC):
    """Test 'aea fetch' in local, remote or mixed mode when it fails."""

    ERROR_MESSAGE = "some error."
    EXPECTED_ERROR_MESSAGE = ""
    MODE = ""

    def _mock_raise_click_exception(self, ctx: Context, *args, **kwargs):
        """Mock 'add_item' so to always fail."""
        raise click.ClickException(BaseTestFetchAgentError.ERROR_MESSAGE)

    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    @pytest.mark.integration
    @mock.patch("aea.cli.fetch.add_item", side_effect=_mock_raise_click_exception)
    @mock.patch("aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_HTTP)
    @mock.patch(
        "aea.cli.fetch.fetch_agent_remote", side_effect=_mock_raise_click_exception
    )
    @mock.patch(
        "aea.cli.registry.fetch.add_item", side_effect=_mock_raise_click_exception
    )
    def test_fetch_negative(self, *_mocks) -> None:
        """Test fetch in mixed mode."""
        if type(self) == BaseTestFetchAgentError:
            pytest.skip("Base test class.")
        with pytest.raises(
            Exception,
            match=self.EXPECTED_ERROR_MESSAGE,
        ):
            self.run_cli_command(
                *(
                    ["-v", "DEBUG", "fetch", str(MY_FIRST_AEA_PUBLIC_ID.to_latest())]
                    + ([self.MODE] if self.MODE else [])
                )
            )


class TestFetchAgentNonMixedErrorLocal(BaseTestFetchAgentError):
    """Test 'aea fetch' in local mode when it fails."""

    EXPECTED_ERROR_MESSAGE = f".*{BaseTestFetchAgentError.ERROR_MESSAGE}"
    MODE = "--local"


class TestFetchAgentMixedModeError(BaseTestFetchAgentError):
    """Test 'aea fetch' in mixed mode when it fails."""

    EXPECTED_ERROR_MESSAGE = f".*{BaseTestFetchAgentError.ERROR_MESSAGE}"
    MODE = ""


class TestFetchAgentRemoteModeError(BaseTestFetchAgentError):
    """Test 'aea fetch' in remote mode when it fails."""

    EXPECTED_ERROR_MESSAGE = rf".*{BaseTestFetchAgentError.ERROR_MESSAGE}"
    MODE = "--remote"


class TestFetchIPFS(BaseAEATestCase):
    """Test fetching package from the IPFS registry."""

    dummy_id = (
        "valory/dummy_agent:bafybeic6plxxhcbokxlce3grbgsm247rgvcb5arwwslm5yv6pt46byr22a"
    )

    def test_run(self) -> None:
        """Test run."""

        with mock.patch(
            "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_IPFS
        ), mock.patch.object(
            IPFSTool,
            "download",
            return_value=self.packages_dir_path
            / "open_aea"
            / "agents"
            / "my_first_aea",
        ), mock.patch(
            "aea.cli.fetch._fetch_agent_deps"
        ) as dep_patch, cd(
            self.t
        ):
            result = self.run_cli_command(
                "fetch",
                self.dummy_id,
                "--remote",
            )

            assert result.exit_code == 0
            dep_patch.assert_called_once()


class TestFetchIPFSFailures(BaseAEATestCase):
    """Test fetching package from the IPFS registry."""

    dummy_id = (
        "valory/dummy_agent:bafybeic6plxxhcbokxlce3grbgsm247rgvcb5arwwslm5yv6pt46byr22a"
    )

    def test_download_failure(self) -> None:
        """Test run."""

        with mock.patch(
            "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_IPFS
        ), mock.patch.object(
            IPFSTool, "download", side_effect=DownloadError
        ), mock.patch.object(
            Path, "exists", return_value=False
        ):
            with pytest.raises(
                click.ClickException,
                match="Error occured while downloading agent",
            ):
                self.run_cli_command(
                    "fetch",
                    self.dummy_id,
                    "--remote",
                )

    def test_os_error(self) -> None:
        """Test run."""

        with mock.patch(
            "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_IPFS
        ), mock.patch.object(IPFSTool, "download", side_effect=shutil.Error):
            with pytest.raises(
                click.ClickException,
            ):
                self.run_cli_command(
                    "fetch",
                    self.dummy_id,
                    "--remote",
                )

    @pytest.mark.skipif(
        condition=(sys.version_info.major <= 3 and sys.version_info.minor <= 7),
        reason="Needs investigation",
    )
    def test_not_an_agent_error(self) -> None:
        """Test run."""

        with TemporaryDirectory() as temp_dir:
            with mock.patch(
                "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_IPFS
            ), mock.patch.object(IPFSTool, "download", return_value=temp_dir):
                with pytest.raises(ClickException, match="is not an agent package"):
                    self.run_cli_command(
                        "fetch",
                        self.dummy_id,
                        "--remote",
                    )


class TestFetchIPFSAlias(BaseAEATestCase):
    """Test fetching package from the IPFS registry."""

    dummy_id = (
        "valory/dummy_agent:bafybeic6plxxhcbokxlce3grbgsm247rgvcb5arwwslm5yv6pt46byr22a"
    )

    def test_alias(self) -> None:
        """Test run."""

        with mock.patch(
            "aea.cli.fetch.get_default_remote_registry", return_value=REMOTE_IPFS
        ), mock.patch.object(
            IPFSTool,
            "download",
            return_value=self.packages_dir_path
            / "open_aea"
            / "agents"
            / "my_first_aea",
        ), mock.patch(
            "aea.cli.fetch._fetch_agent_deps"
        ) as dep_patch:
            result = self.run_cli_command(
                "fetch", self.dummy_id, "--remote", "--alias", "test_alias"
            )

            assert result.exit_code == 0
            assert (self.t / "test_alias").exists()
            dep_patch.assert_called_once()


def test_fetch_mixed_no_local_registry():
    """Test that mixed becomes remote when no local registry."""
    with TemporaryDirectory() as tmp_dir:
        with cd(tmp_dir):
            runner = CliRunner()
            with mock.patch(
                "aea.cli.fetch.fetch_agent_remote"
            ) as fetch_agent_remote_mock:
                result = runner.invoke(
                    cli,
                    ["fetch", "--mixed", "fetchai/my_first_aea"],
                    catch_exceptions=False,
                )
            assert result.exit_code == 0, result.stdout
            fetch_agent_remote_mock.assert_called_once()


def test_fetch_local_no_local_registry():
    """Test that local fetch fails when no local registry."""
    with TemporaryDirectory() as tmp_dir:
        with cd(tmp_dir):
            runner = CliRunner()
            result = runner.invoke(
                cli,
                ["fetch", "--local", "fetchai/my_first_aea"],
                catch_exceptions=False,
            )
            assert result.exit_code == 1, result.stdout
            assert (
                "Registry path not provided and local registry `packages` not found in current (.) and parent directory."
                in result.stdout
            )


def test_fetch_twice_locally():
    """Test fails on fetch if dir exists."""
    with TemporaryDirectory() as tmp_dir:
        with cd(tmp_dir):
            name = "my_first_aea"
            runner = CliRunner()
            result = runner.invoke(
                cli,
                [
                    "--registry-path",
                    PACKAGES_DIR,
                    "fetch",
                    "--local",
                    "fetchai/my_first_aea",
                ],
                catch_exceptions=False,
            )
            assert result.exit_code == 0, result.stdout
            assert os.path.exists(name)

            with pytest.raises(
                ClickException,
                match='Item "my_first_aea" already exists in target folder.',
            ):
                result = runner.invoke(
                    cli,
                    [
                        "--registry-path",
                        PACKAGES_DIR,
                        "fetch",
                        "--local",
                        "fetchai/my_first_aea",
                    ],
                    standalone_mode=False,
                    catch_exceptions=False,
                )


def test_fetch_twice_remote():
    """Test fails on fetch if dir exists."""
    with TemporaryDirectory() as tmp_dir:
        with cd(tmp_dir):
            name = "my_first_aea"
            runner = CliRunner()
            result = runner.invoke(
                cli,
                [
                    "--registry-path",
                    PACKAGES_DIR,
                    "fetch",
                    "--local",
                    "fetchai/my_first_aea",
                ],
                catch_exceptions=False,
            )
            assert result.exit_code == 0, result.stdout
            assert os.path.exists(name)

            with pytest.raises(
                ClickException,
                match='Item "my_first_aea" already exists in target folder.',
            ):
                result = runner.invoke(
                    cli,
                    [
                        "--registry-path",
                        PACKAGES_DIR,
                        "fetch",
                        "--local",
                        "fetchai/my_first_aea",
                    ],
                    standalone_mode=False,
                    catch_exceptions=False,
                )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea list` sub-command."""

import json
import os
import shutil
import tempfile
from pathlib import Path
from unittest import TestCase, mock

import jsonschema
import pytest
from jsonschema import Draft4Validator

from aea.cli import cli

from tests.conftest import (
    AGENT_CONFIGURATION_SCHEMA,
    CLI_LOG_OPTION,
    CONFIGURATION_SCHEMA_DIR,
    CUR_PATH,
    CliRunner,
    MAX_FLAKY_RERUNS,
)
from tests.test_cli.constants import FORMAT_ITEMS_SAMPLE_OUTPUT


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
class TestListProtocols:
    """Test that the command 'aea list protocols' works as expected."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.schema = json.load(open(AGENT_CONFIGURATION_SCHEMA))
        cls.resolver = jsonschema.RefResolver(
            "file://{}/".format(Path(CONFIGURATION_SCHEMA_DIR).absolute()), cls.schema
        )
        cls.validator = Draft4Validator(cls.schema, resolver=cls.resolver)
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        # copy the 'dummy_aea' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "data", "dummy_aea"), Path(cls.t, "dummy_aea"))
        cls.runner = CliRunner()
        os.chdir(Path(cls.t, "dummy_aea"))

        with mock.patch(
            "aea.cli.list.format_items", return_value=FORMAT_ITEMS_SAMPLE_OUTPUT
        ):
            cls.result = cls.runner.invoke(
                cli,
                [*CLI_LOG_OPTION, "list", "protocols"],
                standalone_mode=False,
            )

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        assert self.result.exit_code == 0

    def test_correct_output(self):
        """Test that the command has printed the correct output."""
        compare_text = "{}\n".format(FORMAT_ITEMS_SAMPLE_OUTPUT)
        assert self.result.output == compare_text, self.result.output

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestListConnections:
    """Test that the command 'aea list connections' works as expected."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.schema = json.load(open(AGENT_CONFIGURATION_SCHEMA))
        cls.resolver = jsonschema.RefResolver(
            "file://{}/".format(Path(CONFIGURATION_SCHEMA_DIR).absolute()), cls.schema
        )
        cls.validator = Draft4Validator(cls.schema, resolver=cls.resolver)
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        # copy the 'dummy_aea' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "data", "dummy_aea"), Path(cls.t, "dummy_aea"))
        cls.runner = CliRunner()
        os.chdir(Path(cls.t, "dummy_aea"))

        with mock.patch(
            "aea.cli.list.format_items", return_value=FORMAT_ITEMS_SAMPLE_OUTPUT
        ):
            cls.result = cls.runner.invoke(
                cli, [*CLI_LOG_OPTION, "list", "connections"], standalone_mode=False
            )

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        assert self.result.exit_code == 0

    def test_correct_output(self):
        """Test that the command has printed the correct output."""
        compare_text = "{}\n".format(FORMAT_ITEMS_SAMPLE_OUTPUT)
        assert self.result.output == compare_text, self.result.output

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class TestListSkills:
    """Test that the command 'aea list skills' works as expected."""

    @classmethod
    def setup_class(cls):
        """Set the test up."""
        cls.schema = json.load(open(AGENT_CONFIGURATION_SCHEMA))
        cls.resolver = jsonschema.RefResolver(
            "file://{}/".format(Path(CONFIGURATION_SCHEMA_DIR).absolute()), cls.schema
        )
        cls.validator = Draft4Validator(cls.schema, resolver=cls.resolver)
        cls.cwd = os.getcwd()
        cls.t = tempfile.mkdtemp()
        # copy the 'dummy_aea' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "data", "dummy_aea"), Path(cls.t, "dummy_aea"))
        cls.runner = CliRunner()
        os.chdir(Path(cls.t, "dummy_aea"))

        with mock.patch(
            "aea.cli.list.format_items", return_value=FORMAT_ITEMS_SAMPLE_OUTPUT
        ) as format_items_mock:
            cls.result = cls.runner.invoke(
                cli, [*CLI_LOG_OPTION, "list", "skills"], standalone_mode=False
            )
        format_items_mock.assert_called()

    def test_exit_code_equal_to_zero(self):
        """Assert that the exit code is equal to zero (i.e. success)."""
        assert self.result.exit_code == 0

    def test_correct_output(self):
        """Test that the command has printed the correct output."""
        compare_text = "{}\n".format(FORMAT_ITEMS_SAMPLE_OUTPUT)
        assert self.result.output == compare_text, self.result.output

    @classmethod
    def teardown_class(cls):
        """Tear the test down."""
        os.chdir(cls.cwd)
        try:
            shutil.rmtree(cls.t)
        except (OSError, IOError):
            pass


class ListContractsCommandTestCase(TestCase):
    """Test that the command 'aea list contracts' works as expected."""

    def setUp(self):
        """Set the test up."""
        self.runner = CliRunner()
        self.schema = json.load(open(AGENT_CONFIGURATION_SCHEMA))
        self.resolver = jsonschema.RefResolver(
            "file://{}/".format(Path(CONFIGURATION_SCHEMA_DIR).absolute()), self.schema
        )
        self.validator = Draft4Validator(self.schema, resolver=self.resolver)
        self.cwd = os.getcwd()
        self.t = tempfile.mkdtemp()
        # copy the 'dummy_aea' directory in the parent of the agent folder.
        shutil.copytree(Path(CUR_PATH, "data", "dummy_aea"), Path(self.t, "dummy_aea"))
        os.chdir(Path(self.t, "dummy_aea"))

    @mock.patch("aea.cli.list.list_agent_items")
    @mock.patch("aea.cli.utils.formatting.format_items")
    def test_list_contracts_positive(self, *mocks):
        """Test list contracts command positive result."""
        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "list", "contracts"], standalone_mode=False
        )
        self.assertEqual(result.exit_code, 0)

    def tearDown(self):
        """Tear the test down."""
        os.chdir(self.cwd)
        try:
            shutil.rmtree(self.t)
        except (OSError, IOError):
            pass


class ListAllCommandTestCase(TestCase):
    """Test case for aea list all command."""

    def setUp(self):
        """Set the test up."""
        self.runner = CliRunner()

    @mock.patch("aea.cli.list.list_agent_items", return_value=[])
    @mock.patch("aea.cli.list.format_items")
    @mock.patch("aea.cli.utils.decorators._check_aea_project")
    def test_list_all_no_details_positive(self, *mocks):
        """Test list all command no details positive result."""
        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "list", "all"], standalone_mode=False
        )
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, "")

    @mock.patch("aea.cli.list.list_agent_items", return_value=[{"name": "some"}])
    @mock.patch("aea.cli.list.format_items", return_value="correct")
    @mock.patch("aea.cli.utils.decorators._check_aea_project")
    def test_list_all_positive(self, *mocks):
        """Test list all command positive result."""
        result = self.runner.invoke(
            cli, [*CLI_LOG_OPTION, "list", "all"], standalone_mode=False
        )
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(
            result.output,
            "Connections:\ncorrect\nContracts:\ncorrect\n"
            "Protocols:\ncorrect\nSkills:\ncorrect\n",
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains tests for aea/components/utils.py"""
from itertools import chain
from unittest.mock import patch

from aea.components.utils import _enlist_component_packages, _populate_packages

from tests.test_aea import test_act


def test_modules_enlisted_and_loaded():
    """Test modules enlisted and loaded back."""
    # ensure some packages loaded
    test_act()

    packages = _enlist_component_packages()
    num_of_packages = len(list(chain(*packages.values())))
    assert num_of_packages > 0, "No packages present"

    with patch("aea.components.utils.perform_load_aea_package") as mock_package_load:
        _populate_packages(packages)
        assert mock_package_load.call_count == num_of_packages, packages



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: unixfs.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cunixfs.proto\x12\tunixfs.pb\"\xdc\x01\n\x04\x44\x61ta\x12&\n\x04Type\x18\x01 \x02(\x0e\x32\x18.unixfs.pb.Data.DataType\x12\x0c\n\x04\x44\x61ta\x18\x02 \x01(\x0c\x12\x10\n\x08\x66ilesize\x18\x03 \x01(\x04\x12\x12\n\nblocksizes\x18\x04 \x03(\x04\x12\x10\n\x08hashType\x18\x05 \x01(\x04\x12\x0e\n\x06\x66\x61nout\x18\x06 \x01(\x04\"V\n\x08\x44\x61taType\x12\x07\n\x03Raw\x10\x00\x12\r\n\tDirectory\x10\x01\x12\x08\n\x04\x46ile\x10\x02\x12\x0c\n\x08Metadata\x10\x03\x12\x0b\n\x07Symlink\x10\x04\x12\r\n\tHAMTShard\x10\x05\"\x1c\n\x08Metadata\x12\x10\n\x08MimeType\x18\x01 \x01(\t')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'unixfs_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DATA']._serialized_start=28
  _globals['_DATA']._serialized_end=248
  _globals['_DATA_DATATYPE']._serialized_start=162
  _globals['_DATA_DATATYPE']._serialized_end=248
  _globals['_METADATA']._serialized_start=250
  _globals['_METADATA']._serialized_end=278
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Ledger TX generation and processing benchmark."""
import threading
import time
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Thread
from typing import Dict, List, Optional, Tuple, Union, cast

from aea_cli_benchmark.case_tx_generate.dialogues import (
    FipaDialogues,
    FipaMessage,
    LedgerApiDialogue,
    LedgerApiDialogues,
    SigningDialogues,
)
from aea_cli_benchmark.case_tx_generate.ledger_utils import (
    DEFAULT_FETCH_LEDGER_ADDR,
    DEFAULT_FETCH_LEDGER_REST_PORT,
    DEFAULT_GANACHE_ADDR,
    DEFAULT_GANACHE_CHAIN_ID,
    DEFAULT_GANACHE_PORT,
    FETCHD_CONFIGURATION,
    FUNDED_ETH_PRIVATE_KEY_1,
    GANACHE_CONFIGURATION,
    GAS_PRICE_API_KEY,
    _fetchd_context,
    _ganache_context,
    fund_accounts_from_local_validator,
)
from aea_cli_benchmark.utils import make_agent, make_skill, wait_for_condition

from aea import AEA_DIR as _AEA_DIR
from aea.cli.generate_key import _generate_private_key
from aea.configurations.base import ConnectionConfig, ProtocolConfig
from aea.configurations.data_types import PublicId
from aea.crypto.ledger_apis import ETHEREUM_DEFAULT_CURRENCY_DENOM, LedgerApis
from aea.crypto.registries import make_crypto
from aea.crypto.wallet import Wallet
from aea.helpers.search.models import (
    Constraint,
    ConstraintType,
    Query,
    generate_data_model,
)
from aea.helpers.transaction.base import Terms
from aea.protocols.base import Message, Protocol
from aea.registries.resources import Resources
from aea.skills.base import Handler

from packages.open_aea.protocols.signing.dialogues import SigningDialogue
from packages.open_aea.protocols.signing.message import SigningMessage
from packages.valory.connections.ledger.connection import LedgerConnection
from packages.valory.connections.ledger.connection import (
    PUBLIC_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.valory.protocols.ledger_api.message import LedgerApiMessage


LEDGER_API_ADDRESS = str(LEDGER_CONNECTION_PUBLIC_ID)
AEA_DIR = Path(_AEA_DIR)
PACKAGES_DIR = AEA_DIR.parent / "packages"


class LedgerApiHandler(Handler):
    """Dummy handler to handle messages."""

    SUPPORTED_PROTOCOL = LedgerApiMessage.protocol_id

    def setup(self) -> None:
        """Noop setup."""
        self.context.ledger_api_dialogues = LedgerApiDialogues(
            name=self.skill_id, skill_context=self.context
        )
        self.tx_settled_counter = 0

        self.context.signing_dialogues = SigningDialogues(str(self.skill_id))
        self.event = threading.Event()

    def teardown(self) -> None:
        """Noop teardown."""

    def handle(self, ledger_api_msg: Message) -> None:
        """Handle incoming message."""
        ledger_api_dialogue = cast(
            Optional[LedgerApiDialogue],
            self.context.ledger_api_dialogues.update(ledger_api_msg),
        )
        if ledger_api_msg.performative is LedgerApiMessage.Performative.RAW_TRANSACTION:
            self._handle_raw_transaction(ledger_api_msg, ledger_api_dialogue)
        elif (
            ledger_api_msg.performative
            is LedgerApiMessage.Performative.TRANSACTION_DIGEST
        ):
            self._handle_transaction_digest(ledger_api_msg, ledger_api_dialogue)
        elif (
            ledger_api_msg.performative
            == LedgerApiMessage.Performative.TRANSACTION_RECEIPT
        ):
            self._handle_transaction_receipt(ledger_api_msg, ledger_api_dialogue)
        else:
            raise Exception("BAD ENVELOPE", ledger_api_msg)

    def _handle_raw_transaction(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of raw_transaction performative.

        :param ledger_api_msg: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.context.logger.info("received raw transaction={}".format(ledger_api_msg))
        signing_dialogues = cast(SigningDialogues, self.context.signing_dialogues)
        signing_msg, signing_dialogue = signing_dialogues.create(
            counterparty=self.context.decision_maker_address,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            raw_transaction=ledger_api_msg.raw_transaction,
            terms=ledger_api_dialogue.associated_fipa_dialogue.terms,
        )
        signing_dialogue = cast(SigningDialogue, signing_dialogue)
        signing_dialogue.associated_ledger_api_dialogue = ledger_api_dialogue
        self.context.decision_maker_message_queue.put_nowait(signing_msg)
        self.context.logger.info(
            "proposing the transaction to the decision maker. Waiting for confirmation ..."
        )

    def _handle_transaction_digest(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of transaction_digest performative.

        :param ledger_api_msg: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.context.logger.info(
            "transaction was successfully submitted. Transaction digest={}".format(
                ledger_api_msg.transaction_digest
            )
        )
        ledger_api_msg_ = ledger_api_dialogue.reply(
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
            target_message=ledger_api_msg,
            transaction_digest=ledger_api_msg.transaction_digest,
        )
        self.context.logger.info("checking transaction is settled.")
        self.context.outbox.put_message(message=ledger_api_msg_)

    def _handle_transaction_receipt(
        self, ledger_api_msg: LedgerApiMessage, ledger_api_dialogue: LedgerApiDialogue
    ) -> None:
        """
        Handle a message of transaction_receipt performative.

        :param ledger_api_msg: the ledger api message
        :param ledger_api_dialogue: the ledger api dialogue
        """
        fipa_dialogue = ledger_api_dialogue.associated_fipa_dialogue
        is_settled = LedgerApis.is_transaction_settled(
            fipa_dialogue.terms.ledger_id, ledger_api_msg.transaction_receipt.receipt
        )
        if is_settled:
            self.tx_settled_counter += 1

        self.event.set()


class GenericSigningHandler(Handler):
    """Implement the signing handler."""

    SUPPORTED_PROTOCOL = SigningMessage.protocol_id

    def setup(self) -> None:
        """Implement the setup for the handler."""

    def handle(self, message: Message) -> None:
        """
        Implement the reaction to a message.

        :param message: the message
        """
        signing_msg = cast(SigningMessage, message)

        # recover dialogue
        signing_dialogues = cast(SigningDialogues, self.context.signing_dialogues)
        signing_dialogue = cast(
            Optional[SigningDialogue], signing_dialogues.update(signing_msg)
        )
        if signing_dialogue is None:
            self._handle_unidentified_dialogue(signing_msg)
            return

        # handle message
        if signing_msg.performative is SigningMessage.Performative.SIGNED_TRANSACTION:
            self._handle_signed_transaction(signing_msg, signing_dialogue)
        elif signing_msg.performative is SigningMessage.Performative.ERROR:
            self._handle_error(signing_msg, signing_dialogue)
        else:
            self._handle_invalid(signing_msg, signing_dialogue)

    def teardown(self) -> None:
        """Implement the handler teardown."""

    def _handle_unidentified_dialogue(self, signing_msg: SigningMessage) -> None:
        """
        Handle an unidentified dialogue.

        :param signing_msg: the message
        """
        self.context.logger.info(
            "received invalid signing message={}, unidentified dialogue.".format(
                signing_msg
            )
        )

    def _handle_signed_transaction(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param signing_msg: the signing message
        :param signing_dialogue: the dialogue
        """
        self.context.logger.info("transaction signing was successful.")
        ledger_api_dialogue = signing_dialogue.associated_ledger_api_dialogue
        last_ledger_api_msg = ledger_api_dialogue.last_incoming_message
        if last_ledger_api_msg is None:
            raise ValueError("Could not retrieve last message in ledger api dialogue")
        ledger_api_msg = ledger_api_dialogue.reply(
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            target_message=last_ledger_api_msg,
            signed_transaction=signing_msg.signed_transaction,
        )
        self.context.outbox.put_message(message=ledger_api_msg)
        self.context.logger.info("sending transaction to ledger.")

    def _handle_error(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param signing_msg: the signing message
        :param signing_dialogue: the dialogue
        """
        self.context.logger.info(
            "transaction signing was not successful. Error_code={} in dialogue={}".format(
                signing_msg.error_code, signing_dialogue
            )
        )

    def _handle_invalid(
        self, signing_msg: SigningMessage, signing_dialogue: SigningDialogue
    ) -> None:
        """
        Handle an oef search message.

        :param signing_msg: the signing message
        :param signing_dialogue: the dialogue
        """
        self.context.logger.warning(
            "cannot handle signing message of performative={} in dialogue={}.".format(
                signing_msg.performative, signing_dialogue
            )
        )


class Case:
    """TBenchmark case implementation."""

    LEDGER_TX_SETTLE_TIMEOUT = 20
    QUERY = Query(
        [Constraint("author", ConstraintType("==", "Stephen King"))],
        model=generate_data_model("book_author", {"author": "author of the book"}),
    )

    def __init__(self, ledger_id: str, ledger_api_config: Dict, private_keys: Dict):
        """
        Init case.

        :param ledger_id: str, ledger id, one of fetchai, ethereum
        :param ledger_api_config: config for ledger connection
        :param private_keys: private keys dict to use for wallet contruction
        """
        self.ledger_id = ledger_id
        self.ledger_api_config = ledger_api_config
        self.agent_name = "Agent"
        self.skill_public_id = PublicId.from_str("benchmark/test_skill:0.1.0")
        self.private_keys = private_keys

    @property
    def ledger_handler(self) -> LedgerApiHandler:
        """Get ledger api handler instance."""
        return self.skill.handlers["ledger_handler"]

    @property
    def tx_settled_counter(self) -> int:
        """Get amount of txs settled."""
        return self.ledger_handler.tx_settled_counter

    def wait_tx_settled(self) -> None:
        """Wait for tx settled."""
        self.ledger_handler.event.wait(self.LEDGER_TX_SETTLE_TIMEOUT)
        self.ledger_handler.event.clear()

    @property
    def ledger_api_dialogues(self) -> LedgerApiDialogues:
        """Get ledger api dialogues."""
        return self.skill.skill_context.ledger_api_dialogues

    @property
    def my_addr(self) -> str:
        """Get my agent address."""
        return self.skill.skill_context.agent_address

    def make_ledger_msg(
        self, sender_address: str, counterparty_address: str
    ) -> LedgerApiMessage:
        """Make ledger api message to be signed and published over ledger netework."""
        _, dialogue = self.fipa_dialogues.create(
            counterparty="some",
            performative=FipaMessage.Performative.CFP,
            query=self.QUERY,
        )
        dialogue.terms = Terms(
            ledger_id=self.ledger_id,
            sender_address=sender_address,
            counterparty_address=counterparty_address,
            amount_by_currency_id={"FET": -2},
            quantities_by_good_id={"test": 10},
            is_sender_payable_tx_fee=True,
            nonce="",
            fee_by_currency_id={"FET": 45000},
        )

        ledger_api_msg, ledger_api_dialogue = self.ledger_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
            terms=dialogue.terms,
        )
        ledger_api_dialogue.associated_fipa_dialogue = dialogue
        return ledger_api_msg

    def start_agent(self) -> None:
        """Construct and start agent."""
        wallet = Wallet(self.private_keys, self.private_keys)
        resources = Resources()
        pconfig = ProtocolConfig(
            name=LedgerApiMessage.protocol_id.name,
            author=LedgerApiMessage.protocol_id.author,
            version=LedgerApiMessage.protocol_id.version,
            protocol_specification_id=LedgerApiMessage.protocol_specification_id,
        )

        resources.add_protocol(
            Protocol(
                pconfig,
                message_class=LedgerApiMessage,
            )
        )

        connection = LedgerConnection(
            data_dir="./",
            configuration=ConnectionConfig(
                connection_id=LedgerConnection.connection_id,
                **{"ledger_apis": self.ledger_api_config},
            ),
            identity=None,
        )
        resources.add_connection(connection)

        self.agent = make_agent(
            agent_name=self.agent_name,
            resources=resources,
            wallet=wallet,
            packages_dir=str(PACKAGES_DIR),
        )

        skill = make_skill(
            self.agent,
            handlers={
                "ledger_handler": LedgerApiHandler,
                "sign_handler": GenericSigningHandler,
            },
            skill_id=self.skill_public_id,
        )
        self.skill = skill
        self.fipa_dialogues = FipaDialogues(
            name=skill.public_id, skill_context=skill.skill_context
        )
        skill.skill_context.agent_address
        self.agent.resources.add_skill(skill)
        self.thread = Thread(target=self.agent.start, daemon=True)
        self.thread.start()
        wait_for_condition(lambda: self.agent.is_running, timeout=5)

    def stop_agent(self) -> None:
        """Stop agent."""
        self.agent.runtime.stop()
        wait_for_condition(lambda: not self.agent.is_running, timeout=20)
        self.thread.join(60)

    def put_message_and_wait(self, msg: Message) -> None:
        """Put ledger api message and wait tx constructed, signed and settled."""
        self.agent.outbox.put_message(msg)
        self.wait_tx_settled()

    def run(self, time_in_seconds: float) -> Tuple[int, float]:
        """Run a test case."""
        self.start_agent()

        if self.ledger_id == "fetchai":
            fund_accounts_from_local_validator([self.my_addr], 10000000000000000000)

        target_addr = make_crypto(self.ledger_id).address
        start_time = time.time()
        execution_time = 0

        while execution_time < time_in_seconds:
            msg = self.make_ledger_msg(self.my_addr, target_addr)
            self.put_message_and_wait(msg)
            counted = self.tx_settled_counter
            execution_time = time.time() - start_time

        return counted, execution_time


def run(ledger_id: str, running_time: float) -> List[Tuple[str, Union[int, float]]]:
    """Check tx processing speed."""
    # pylint: disable=import-outside-toplevel,unused-import
    # import manually due to some lazy imports in decision_maker
    import aea.decision_maker.default  # noqa: F401

    with TemporaryDirectory() as tmp_dir:
        private_key_path = Path(tmp_dir) / "priv.key"

        if ledger_id == "fetchai":
            ctx = _fetchd_context(FETCHD_CONFIGURATION, timeout=5)
            ledger_api_config = {
                "fetchai": {
                    "address": f"{DEFAULT_FETCH_LEDGER_ADDR}:{DEFAULT_FETCH_LEDGER_REST_PORT}",
                    "chain_id": FETCHD_CONFIGURATION["chain_id"],
                    "denom": FETCHD_CONFIGURATION["denom"],
                }
            }
            private_keys = private_keys = _generate_private_key(
                ledger_id, private_key_path
            )
        elif ledger_id == "ethereum":
            ctx = _ganache_context(
                GANACHE_CONFIGURATION,
                DEFAULT_GANACHE_ADDR,
                DEFAULT_GANACHE_PORT,
                timeout=5,
            )
            ledger_api_config = {
                "ethereum": {
                    "address": f"{DEFAULT_GANACHE_ADDR}:{DEFAULT_GANACHE_PORT}",
                    "chain_id": DEFAULT_GANACHE_CHAIN_ID,
                    "denom": ETHEREUM_DEFAULT_CURRENCY_DENOM,
                    "gas_price_api_key": GAS_PRICE_API_KEY,
                }
            }
            with open(private_key_path, "w") as f:
                f.write(FUNDED_ETH_PRIVATE_KEY_1)
            private_keys = {"ethereum": private_key_path}
        else:
            raise ValueError(f"Bad ledger id {ledger_id}")

        with ctx:
            case = Case(
                ledger_id=ledger_id,
                ledger_api_config=ledger_api_config,
                private_keys=private_keys,
            )

            tx_processed, execution_time = case.run(
                time_in_seconds=running_time,
            )

        return [
            ("run_time (seconds)", execution_time),
            ("tx process speed (seconds)", execution_time / tx_processed),
            ("tx processed", tx_processed),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Hypothesis strategies for data types."""

import collections
import re

import semver
from hypothesis import strategies as st
from packaging.specifiers import Specifier, SpecifierSet

from aea.configurations.data_types import (
    ComponentId,
    ComponentType,
    Dependency,
    GitRef,
    PackageId,
    PackageType,
    PackageVersion,
    PublicId,
    PyPIPackageName,
)
from aea.helpers.base import IPFSHash, SimpleId


positive_integer_strategy = st.integers(min_value=0)
user_string_pattern = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]{0,127}")
user_string_strategy = st.from_regex(user_string_pattern, fullmatch=True)
simple_id_strategy = st.from_regex(SimpleId.REGEX, fullmatch=True)
ipfs_hash_strategy = st.from_regex(IPFSHash.REGEX, fullmatch=True)
pypi_package_name_strategy = st.from_regex(PyPIPackageName.REGEX, fullmatch=True)
gitref_strategy = st.from_regex(GitRef.REGEX, fullmatch=True)
specifier_strategy = st.from_regex(Specifier._regex, fullmatch=True)


st.register_type_strategy(collections.UserString, user_string_strategy)
st.register_type_strategy(SimpleId, simple_id_strategy)
st.register_type_strategy(IPFSHash, ipfs_hash_strategy)
st.register_type_strategy(PyPIPackageName, pypi_package_name_strategy)
st.register_type_strategy(GitRef, gitref_strategy)
st.register_type_strategy(Specifier, specifier_strategy)


version_info_strategy = st.builds(
    lambda kwargs: semver.VersionInfo(**kwargs),
    st.fixed_dictionaries(
        dict(
            major=positive_integer_strategy,
            minor=positive_integer_strategy,
            patch=positive_integer_strategy,
        ),
    ),
)
st.register_type_strategy(semver.VersionInfo, version_info_strategy)


package_version_strategy = st.builds(
    lambda kwargs: PackageVersion(**kwargs),
    st.fixed_dictionaries(dict(version_like=version_info_strategy)),
)
st.register_type_strategy(PackageVersion, package_version_strategy)


s = st.fixed_dictionaries(
    dict(
        author=st.from_type(SimpleId),
        name=st.from_type(SimpleId),
        version=st.from_type(semver.VersionInfo),
        package_hash=st.from_type(IPFSHash),
    )
)
public_id_strategy = st.builds(lambda kwargs: PublicId(**kwargs), s)
st.register_type_strategy(PublicId, public_id_strategy)


s = st.fixed_dictionaries(
    dict(
        package_type=st.from_type(PackageType),
        public_id=st.from_type(PublicId),
    )
)
package_id_strategy = st.builds(lambda kwargs: PackageId(**kwargs), s)
st.register_type_strategy(PackageId, package_id_strategy)


s = st.fixed_dictionaries(
    dict(
        component_type=st.from_type(ComponentType),
        public_id=st.from_type(PublicId),
    )
)
component_id_strategy = st.builds(lambda kwargs: ComponentId(**kwargs), s)
st.register_type_strategy(ComponentId, component_id_strategy)


specifier_set_strategy = st.builds(
    lambda ss: ",".join(ss), st.lists(specifier_strategy)
)
s = st.fixed_dictionaries(dict(specifiers=specifier_set_strategy))
specifier_set_strategy = st.builds(lambda kwargs: SpecifierSet(**kwargs), s)
st.register_type_strategy(SpecifierSet, specifier_set_strategy)


s = st.fixed_dictionaries(
    dict(
        name=st.from_type(PyPIPackageName),
        version=st.from_type(SpecifierSet),
        index=st.text(),
        git=st.text(),
        ref=st.from_type(GitRef),
    )
)
dependency_strategy = st.builds(lambda kwargs: Dependency(**kwargs), s)
st.register_type_strategy(Dependency, dependency_strategy)


del s  # clean-up namespace



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains helper methods and classes for the 'aea' package."""
import codecs
import hashlib
import io
import os
import platform
import re
from pathlib import Path
from typing import Any, Dict, Generator, Optional, Sized, Tuple, cast

import base58

from aea.helpers.cid import to_v1
from aea.helpers.io import open_file
from aea.helpers.ipfs.utils import _protobuf_python_implementation


# https://github.com/multiformats/multicodec/blob/master/table.csv
SHA256_ID = "12"  # 0x12
LEN_SHA256 = "20"  # 0x20


with _protobuf_python_implementation():  # pylint: disable=import-outside-toplevel
    from aea.helpers.ipfs.pb import merkledag_pb2, unixfs_pb2
    from aea.helpers.ipfs.pb.merkledag_pb2 import PBNode  # type: ignore


def _is_text(file_path: str) -> bool:
    """Check if a file can be read as text or not."""
    try:
        with open_file(file_path, "r") as f:
            f.readline()
        return True
    except UnicodeDecodeError:
        return False


def _dos2unix(file_content: bytes) -> bytes:
    """
    Replace occurrences of Windows line terminator CR/LF with only LF.

    :param file_content: the content of the file.
    :return: the same content but with the line terminator
    """
    return re.sub(b"\r\n", b"\n", file_content, flags=re.M)


def _read(file_path: str) -> bytes:
    """Read and verify the file is not empty."""
    with open(file_path, "rb") as file:
        data = file.read()
    if len(data) == 0:
        raise ValueError(f"File cannot be empty: {file_path}")
    if platform.system() == "Windows" and _is_text(file_path=file_path):
        data = _dos2unix(data)
    return data


def chunks(data: Sized, size: int) -> Generator:
    """Yield successivesize chunks from data."""
    for i in range(0, len(data), size):
        yield data[i : i + size]  # type: ignore


class IPFSHashOnly:
    """A helper class which allows construction of an IPFS hash without interacting with an IPFS daemon."""

    DEFAULT_CHUNK_SIZE = 262144
    # according to https://pkg.go.dev/github.com/ipfs/go-ipfs-chunker#pkg-constants

    @classmethod
    def get(cls, file_path: str, wrap: bool = True, cid_v1: bool = True) -> str:
        """Get the IPFS hash."""
        if os.path.isdir(file_path):
            return cls.hash_directory(file_path, wrap=wrap, cid_v1=cid_v1)

        return cls.hash_file(file_path, wrap=wrap, cid_v1=cid_v1)

    @classmethod
    def hash_file(cls, file_path: str, wrap: bool = True, cid_v1: bool = True) -> str:
        """
        Get the IPFS hash for a single file.

        :param file_path: the file path
        :param wrap: whether to wrap the content in wrapper node or not
        :param cid_v1: whether to use CID v1 hashes
        :return: the ipfs hash
        """
        data = _read(file_path)
        return cls.hash_bytes(
            data=data, wrap=wrap, cid_v1=cid_v1, file_name_if_wrap=Path(file_path).name
        )

    @classmethod
    def hash_bytes(
        cls,
        data: bytes,
        wrap: bool = True,
        cid_v1: bool = True,
        file_name_if_wrap: Optional[str] = None,
    ) -> str:
        """
        Get the IPFS hash for a single file.

        :param data: bytes
        :param wrap: whether to wrap the content in wrapper node or not
        :param cid_v1: whether to use CID v1 hashes
        :param file_name_if_wrap: optional str with filename applied if wrap is True
        :return: the ipfs hash
        """
        file_pb, file_length = cls._pb_serialize_bytes(data)

        if wrap:
            if not file_name_if_wrap:
                raise ValueError(
                    "`file_name_if_wrap` is required if wrap option is True"
                )
            link_hash = cls._generate_multihash_bytes(file_pb)
            link = cls.create_link(link_hash, file_length, file_name_if_wrap)
            file_hash = cls.wrap_in_a_node(link)
        else:
            file_hash = cls._generate_multihash(file_pb)

        if cid_v1:
            return to_v1(file_hash)

        return file_hash

    @classmethod
    def hash_directory(
        cls, dir_path: str, wrap: bool = True, cid_v1: bool = True
    ) -> str:
        """
        Get the IPFS hash for a directory.

        :param dir_path: the directory path
        :param wrap: whether to wrap the content in wrapper node or not
        :param cid_v1: whether to use CID v1 hashes
        :return: the ipfs hash
        """

        path = Path(dir_path)
        hashed_dir = cls._hash_directory_recursively(path)

        if wrap:
            link = cls.create_link(
                cast(bytes, hashed_dir.get("hash_bytes")),
                len(cast(bytes, hashed_dir.get("serialization")))
                + cast(int, hashed_dir.get("content_size")),
                path.name,
            )
            dir_hash = cls.wrap_in_a_node(link)
        else:
            dir_hash = cast(str, hashed_dir.get("hash"))

        if cid_v1:
            return to_v1(dir_hash)

        return dir_hash

    @staticmethod
    def create_link(link_hash: bytes, tsize: int, name: str) -> Any:
        """Create PBLink object."""
        link = merkledag_pb2.PBLink()  # type: ignore # pylint: disable=no-member
        link.Hash = link_hash
        link.Tsize = tsize
        link.Name = name

        return link

    @classmethod
    def wrap_in_a_node(cls, link: Any) -> str:
        """Wrap content in a wrapper node."""
        wrapper_node = PBNode()
        wrapper_node.Links.append(link)  # type: ignore # pylint: disable=no-member

        wrapper_node_data = unixfs_pb2.Data()  # type: ignore # pylint: disable=no-member
        # type: ignore  # pylint: disable=no-member
        wrapper_node_data.Type = unixfs_pb2.Data.Directory  # type: ignore
        wrapper_node.Data = wrapper_node_data.SerializeToString(
            deterministic=True
        )  # type: ignore # pylint: disable=no-member

        wrapper_node_serialization = cls._serialize(wrapper_node)
        wrapper_node_hash_bytes = cls._generate_multihash_bytes(
            wrapper_node_serialization
        )

        return base58.b58encode(wrapper_node_hash_bytes).decode()

    @classmethod
    def _hash_directory_recursively(cls, root: Path) -> Dict:
        """Hash directories recursively, starting from provided root directory."""

        root_node = PBNode()
        content_size = 0

        for child_path in sorted(root.iterdir(), key=lambda x: x.name):
            if child_path.is_dir():
                if child_path.name == "__pycache__":
                    continue
                metadata = cls._hash_directory_recursively(child_path)
                content_size_child = len(
                    cast(bytes, metadata.get("serialization"))
                ) + cast(int, metadata.get("content_size"))
                root_node.Links.append(  # type: ignore # pylint: disable=no-member
                    cls.create_link(
                        cast(bytes, metadata.get("hash_bytes")),
                        content_size_child,
                        child_path.name,
                    )
                )
                content_size += content_size_child

            else:
                if child_path.name.endswith(".pyc"):
                    continue
                data = _read(str(child_path))
                file_pb, file_length = cls._pb_serialize_bytes(data)
                child_hash = cls._generate_multihash_bytes(file_pb)
                content_size += file_length
                root_node.Links.append(  # type: ignore # pylint: disable=no-member
                    cls.create_link(child_hash, file_length, child_path.name)
                )

        root_node_data = unixfs_pb2.Data()  # type: ignore # pylint: disable=no-member
        # type: ignore  # pylint: disable=no-member
        root_node_data.Type = unixfs_pb2.Data.Directory  # type: ignore
        root_node.Data = root_node_data.SerializeToString(
            deterministic=True
        )  # type: ignore # pylint: disable=no-member

        root_node_serialization = cls._serialize(root_node)
        root_node_hash_bytes = cls._generate_multihash_bytes(root_node_serialization)
        root_node_hash = base58.b58encode(root_node_hash_bytes).decode()

        return {
            "serialization": root_node_serialization,
            "hash_bytes": root_node_hash_bytes,
            "hash": root_node_hash,
            "content_size": content_size,
        }

    @classmethod
    def _make_unixfs_pb2(cls, data: bytes) -> bytes:
        if len(data) > cls.DEFAULT_CHUNK_SIZE:  # pragma: nocover
            raise ValueError("Data is too big! use chunks!")
        data_pb = unixfs_pb2.Data()  # type: ignore  # pylint: disable=no-member
        data_pb.Type = unixfs_pb2.Data.File  # type: ignore # pylint: disable=no-member
        data_pb.Data = data
        data_pb.filesize = len(data)
        serialized_data = data_pb.SerializeToString(deterministic=True)
        return serialized_data

    @classmethod
    def _pb_serialize_data(cls, data: bytes) -> bytes:
        outer_node = PBNode()  # type: ignore
        outer_node.Data = cls._make_unixfs_pb2(data)
        result = cls._serialize(outer_node)
        return result

    @classmethod
    def _serialize(cls, pb_node: PBNode) -> bytes:  # type: ignore
        """Serialize PBNode instance with fixed fields sequence."""
        f = io.BytesIO()
        # type: ignore
        for field_descriptor, field_value in reversed(pb_node.ListFields()):
            field_descriptor._encoder(  # pylint: disable=protected-access
                f.write, field_value, True
            )
        return f.getvalue()

    @classmethod
    def _pb_serialize_bytes(cls, data: bytes) -> Tuple[bytes, int]:
        """
        Serialize a bytes object..

        :param data: a bytes string
        :return: a bytes string in protobuf serialization, content size
        """
        if len(data) > cls.DEFAULT_CHUNK_SIZE:
            content_size = 0
            outer_node = PBNode()  # type: ignore
            data_pb = unixfs_pb2.Data()  # type: ignore   # pylint: disable=no-member
            data_pb.Type = unixfs_pb2.Data.File  # type: ignore  # pylint: disable=no-member
            data_pb.filesize = len(data)
            for chunk in chunks(data, cls.DEFAULT_CHUNK_SIZE):
                block = cls._pb_serialize_data(chunk)
                block_length = len(block)
                content_size += block_length
                outer_node.Links.append(  # type: ignore # pylint: disable=no-member
                    cls.create_link(
                        cls._generate_multihash_bytes(block), block_length, ""
                    )
                )
                # type: ignore # pylint: disable=no-member
                data_pb.blocksizes.append(len(chunk))
            outer_node.Data = data_pb.SerializeToString(deterministic=True)
            file_pb = cls._serialize(outer_node)
            content_size += len(file_pb)
            return file_pb, content_size

        file_pb = cls._pb_serialize_data(data)
        return file_pb, len(file_pb)

    @staticmethod
    def _generate_multihash_bytes(pb_data: bytes) -> bytes:
        sha256_hash = hashlib.sha256(pb_data).hexdigest()
        multihash_hex = SHA256_ID + LEN_SHA256 + sha256_hash
        multihash_bytes = codecs.decode(str.encode(multihash_hex), "hex")
        return cast(bytes, multihash_bytes)

    @classmethod
    def _generate_multihash(cls, pb_data: bytes) -> str:
        """
        Generate an IPFS multihash.

        Uses the default IPFS hashing function: sha256

        :param pb_data: the data to be hashed
        :return: string representing the hash
        """
        multihash_bytes = cls._generate_multihash_bytes(pb_data)
        ipfs_hash = base58.b58encode(multihash_bytes)
        return ipfs_hash.decode()

    @classmethod
    def _generate_hash(cls, data: bytes) -> str:
        """Generate hash for data."""
        pb_data, _ = cls._pb_serialize_bytes(data)
        return cls._generate_multihash(pb_data)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the environment variables support."""
import json
import re
from collections.abc import Mapping as MappingType
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union, cast

from aea.configurations.data_types import PublicId
from aea.helpers.constants import (
    FALSE_EQUIVALENTS,
    FROM_STRING_TO_TYPE,
    JSON_TYPES,
    NULL_EQUIVALENTS,
)


ENV_VARIABLE_RE = re.compile(r"^\$\{(([A-Z0-9_]+):?)?([a-z]+)?(:(.+))?}$")


def is_env_variable(value: Any) -> bool:
    """Check is variable string with env variable pattern."""
    return isinstance(value, str) and bool(ENV_VARIABLE_RE.match(value))


def export_path_to_env_var_string(export_path: List[str]) -> str:
    """Conver export path to environment variable string."""
    env_var_string = "_".join(map(str, export_path))
    return env_var_string.upper()


NotSet = object()


def parse_list(var_prefix: str, env_variables: dict) -> str:
    """Parse list object."""
    values = {}
    _vars = list(filter(lambda x: x.startswith(var_prefix), env_variables.keys()))
    for var in _vars:
        _, idx, *sub_var = var.replace(var_prefix, "").split("_")
        if len(sub_var) > 0:
            values[idx] = json.loads(
                parse_list(
                    var_prefix=f"{var_prefix}_{idx}",
                    env_variables=env_variables,
                )
            )
            continue
        try:
            values[idx] = json.loads(str(env_variables[var]))
        except (json.JSONDecodeError, ValueError):
            values[idx] = env_variables[var]
    if all(map(lambda x: isinstance(json.loads(x), int), values.keys())):
        return json.dumps([values[idx] for idx in sorted(values)])
    return json.dumps({json.loads(key): val for key, val in values.items()})


def replace_with_env_var(
    value: str,
    env_variables: dict,
    default_value: Any = NotSet,
    default_var_name: Optional[str] = None,
) -> JSON_TYPES:
    """Replace env var with value."""
    result = ENV_VARIABLE_RE.match(value)

    if not result:
        return value

    _, var_name, type_str, _, default = result.groups()
    if var_name is None and default_var_name is not None:
        var_name = default_var_name

    if var_name in env_variables:
        var_value = env_variables[var_name]
    elif type_str == "list":
        var_value = parse_list(
            var_prefix=var_name,
            env_variables=env_variables,
        )
        var_value = (default or var_value) if var_value == "[]" else var_value
    elif default is not None:
        var_value = default
    elif default_value is not NotSet:
        var_value = default_value
    else:
        raise ValueError(
            f"`{var_name}` not found in env variables and no default value set! Please ensure a .env file is provided."
        )

    if type_str is not None:
        var_value = convert_value_str_to_type(var_value, type_str)
    return var_value


def apply_env_variables(
    data: Union[Dict, List[Dict]],
    env_variables: Mapping[str, Any],
    path: Optional[List[str]] = None,
    default_value: Any = NotSet,
) -> JSON_TYPES:
    """Create new resulting dict with env variables applied."""
    path = path or []

    if isinstance(data, (list, tuple)):
        result = []
        for i, obj in enumerate(data):
            result.append(
                apply_env_variables(
                    data=obj,
                    env_variables=env_variables,
                    path=[*path, str(i)],
                    default_value=default_value,
                )
            )
        return result

    if isinstance(data, MappingType):
        return {
            k: apply_env_variables(
                data=v,
                env_variables=env_variables,
                path=[*path, k],
                default_value=default_value,
            )
            for k, v in data.items()
        }

    if is_env_variable(data):
        return replace_with_env_var(
            data,
            env_variables,
            default_value,
            default_var_name=export_path_to_env_var_string(export_path=path),
        )

    return data


def convert_value_str_to_type(value: str, type_str: str) -> JSON_TYPES:
    """Convert value by type name to native python type."""
    try:
        type_ = FROM_STRING_TO_TYPE[type_str]
        if type_ == bool:
            return value not in FALSE_EQUIVALENTS
        if type_ is None or value in NULL_EQUIVALENTS:
            return None
        if type_ in (dict, list):
            return json.loads(value)
        return type_(value)
    except (ValueError, json.decoder.JSONDecodeError):
        raise ValueError(f"Cannot convert string `{value}` to type `{type_.__name__}`")  # type: ignore
    except KeyError:
        raise KeyError(f"`{type_str}` is not a valid python data type")


def apply_env_variables_on_agent_config(
    data: List[Dict],
    env_variables: Mapping[str, Any],
) -> List[Dict]:
    """Create new resulting dict with env variables applied."""

    agent_config, *overrides = data
    agent_config_new = apply_env_variables(
        data=agent_config,
        env_variables=env_variables,
        path=[
            "agent",
        ],
    )

    overrides_new = []
    for component_config in overrides:
        component_name = PublicId.from_str(
            cast(str, component_config.get("public_id")),
        ).name
        component_type = cast(str, component_config.get("type"))
        new_component_config = cast(
            Dict,
            apply_env_variables(
                data=component_config,
                env_variables=env_variables,
                path=[
                    component_type,
                    component_name,
                ],
            ),
        )
        overrides_new.append(new_component_config)

    return [cast(Dict, agent_config_new), *overrides_new]


def is_strict_list(data: Union[List, Tuple]) -> bool:
    """
    Check if a data list is an strict list

    The data list contains a mapping object we need to process it as an
    object containing configurable parameters. For example

    cert_requests:
      - public_key: example_public_key

    This will get exported as `CONNECTION_NAME_CERT_REQUESTS_0_PUBLIC_KEY=example_public_key`

    Where as

    parameters:
     - hello
     - world

     will get exported as `SKILL_NAME_PARAMETERS=["hello", "world"]`

    :param data: Data list
    :return: Boolean specifying whether it's a strict list or not
    """
    is_strict = True
    for obj in data:
        if isinstance(obj, dict):
            return False
        if isinstance(obj, (list, tuple)):
            if not is_strict_list(data=obj):
                return False
    return is_strict


def generate_env_vars_recursively(
    data: Union[Dict, List],
    export_path: List[str],
) -> Dict:
    """Generate environment variables recursively."""
    env_var_dict = {}

    if isinstance(data, dict):
        for key, value in data.items():
            env_var_dict.update(
                generate_env_vars_recursively(
                    data=value,
                    export_path=[*export_path, key],
                )
            )
    elif isinstance(data, list):
        if is_strict_list(data=data):
            env_var_dict[
                export_path_to_env_var_string(export_path=export_path)
            ] = json.dumps(data, separators=(",", ":"))
        else:
            for key, value in enumerate(data):
                env_var_dict.update(
                    generate_env_vars_recursively(
                        data=value,
                        export_path=[*export_path, key],
                    )
                )
    else:
        env_var_dict[export_path_to_env_var_string(export_path=export_path)] = data

    return env_var_dict



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Check amount of time and mem for agent setup."""
import os
import time
from pathlib import Path
from statistics import mean
from tempfile import TemporaryDirectory
from typing import List, Tuple, Union

from aea_cli_benchmark.utils import get_mem_usage_in_mb
from click.testing import CliRunner

from aea import AEA_DIR as _AEA_DIR
from aea.aea_builder import AEABuilder


AEA_DIR = Path(_AEA_DIR)
PACKAGES_DIR = AEA_DIR.parent / "packages"


def run(agents: int) -> List[Tuple[str, Union[int, float]]]:
    """Check construction time and memory usage."""
    from aea.cli.core import cli

    load_times = []
    full_times = []
    with TemporaryDirectory() as tmp_dir:
        os.chdir(tmp_dir)
        if (
            CliRunner()
            .invoke(
                cli,
                [
                    f"--registry-path={PACKAGES_DIR}",
                    "fetch",
                    "--local",
                    "fetchai/my_first_aea",
                    "--alias",
                    "agent",
                ],
                catch_exceptions=False,
            )
            .exit_code
            != 0
        ):
            raise Exception("fetch failed")
        agent_dir = Path(tmp_dir) / "agent"
        os.chdir(agent_dir)
        if (
            CliRunner()
            .invoke(cli, ["generate-key", "fetchai"], catch_exceptions=False)
            .exit_code
            != 0
        ):
            raise Exception("generate-key failed")
        if (
            CliRunner()
            .invoke(cli, ["add-key", "fetchai"], catch_exceptions=False)
            .exit_code
            != 0
        ):
            raise Exception("add-key failed")
        agents_list = []
        env_mem_usage = get_mem_usage_in_mb()
        for _ in range(agents):
            start_time = time.time()
            builder = AEABuilder.from_aea_project(agent_dir)
            load_times.append(time.time() - start_time)
            agents_list.append(builder.build())
            full_times.append(time.time() - start_time)
        mem_usage = get_mem_usage_in_mb()

    return [
        ("avg config load time", mean(load_times)),
        ("avg full construction", mean(full_times)),
        ("avg build time", mean(full_times) - mean(load_times)),
        ("agent mem usage (Mb)", mem_usage - env_mem_usage),
    ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Module with helpers constants."""
from typing import Dict, List, Union


FALSE_EQUIVALENTS = ["f", "false", "False", "0"]
NULL_EQUIVALENTS = ["Null", "null", "None", "none"]
FROM_STRING_TO_TYPE = dict(
    str=str,
    int=int,
    bool=bool,
    float=float,
    dict=dict,
    list=list,
    none=None,
)
JSON_TYPES = Union[Dict, str, List, None, int, float]

NETWORK_REQUEST_DEFAULT_TIMEOUT = 60.0  # in seconds



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Parser and plotter for profiling logs. Pyqt5 might be required in some systems in order for the script to work: pip install PyQt5"""
import datetime
import json
import os
import random
import re
from enum import Enum
from typing import Dict, List, Optional

from matplotlib import pyplot as plt  # type: ignore


LOG_FILE_DIR_BASE = "/tmp/oracle_logs"  # nosec


class TimeFormat(Enum):
    """Time formats"""

    RELATIVE = "relative"
    ABSOLUTE = "absolute"


class LineStyle(Enum):
    """Line styles"""

    SOLID = "-"
    DASHED = "--"
    DOTTED = ":"
    DASH_DOTTED = "-."


class LineColor(Enum):
    """Line colors"""

    BLACK = "k"
    RED = "r"
    GREEN = "g"
    BLUE = "b"
    YELLOW = "y"
    CYAN = "c"
    MAGENTA = "m"


def get_latest_log_by_agent_index(agent_index: Optional[int] = 0):
    """Retrieve the last log for the specified agent in the log directory"""
    most_recent_logs_dir = sorted(os.listdir(LOG_FILE_DIR_BASE))[-1]
    log_file_path = f"{LOG_FILE_DIR_BASE}/{most_recent_logs_dir}/abci{agent_index}.log"
    return log_file_path


class LogParser:
    """A class to parse and plot information from agent logs"""

    def __init__(
        self,
        log_file_path: Optional[str] = None,
        agent_index: Optional[int] = 0,
        time_format: Optional[TimeFormat] = TimeFormat.RELATIVE,
    ) -> None:
        """Log parser"""
        self.log_file_path = (
            log_file_path
            if log_file_path
            else get_latest_log_by_agent_index(agent_index)
        )
        self.agent_index = agent_index
        self.time_format = time_format
        self.default_x_label = (
            "Time [min]" if self.time_format == TimeFormat.RELATIVE else "Time"
        )
        self.time_origin: Optional[datetime.datetime] = None
        self.line_trackers: Dict[str, Dict] = {}
        self.figures: List = []

    def add_tracker(
        self,
        tracker_name: str,
        regex: str,
        tracker_type: str = "line",
        figure_names: Optional[List[str]] = None,
        line_styles: Optional[dict] = None,
    ) -> None:
        """Add a line of interest in the log to track"""

        # Read all the variables in the regex and initialize them
        var_name_regex = r"\(\?P<(?P<var_name>.*?)>"
        var_names = re.findall(var_name_regex, regex)
        if tracker_type == "event":
            var_names = ["event"]
        var_data: Dict[str, Dict] = {
            var_name: {"times": [], "values": [], "line": {}} for var_name in var_names
        }

        for var_name in var_names:
            line_style = (
                line_styles[var_name]["style"]
                if line_styles and var_name in line_styles
                else random.choice([e.value for e in LineColor])  # nosec
            )
            line_color = (
                line_styles[var_name]["color"]
                if line_styles and var_name in line_styles
                else random.choice([e.value for e in LineStyle])  # nosec
            )
            var_data[var_name]["line_style"] = f"{line_color}{line_style}"
            var_data[var_name]["line_width"] = (
                line_styles[var_name]["width"]
                if line_styles and var_name in line_styles
                else 1
            )

        self.line_trackers[tracker_name] = {
            "regex": regex,
            "type": tracker_type,
            "figure_names": figure_names if figure_names else [tracker_name],
            "var_data": var_data,
        }

    def process(self) -> None:
        """Process the log"""
        with open(self.log_file_path, "r") as log_file:
            content = log_file.readlines()

            # Guess log type
            LOG_TYPE = ""
            try:
                json.loads(content[0])
            except json.JSONDecodeError:
                LOG_TYPE = "text"
            else:
                LOG_TYPE = "json"

            for line in content:
                line_time: Optional[datetime.datetime] = None
                line_data: str = line

                # Handle json logs
                if LOG_TYPE == "json":
                    line_json = json.loads(line)
                    line_time = datetime.datetime.strptime(
                        line_json["time"][:-4], "%Y-%m-%dT%H:%M:%S.%f"
                    )
                    line_data = line_json["log"]
                    if not self.time_origin:
                        self.time_origin = line_time

                # Add line data to corresponding plots
                for tracker_name, tracker_data in self.line_trackers.items():
                    match = re.match(tracker_data["regex"], line_data)
                    if match:
                        if tracker_data["type"] == "event":
                            self.line_trackers[tracker_name]["var_data"]["event"][
                                "times"
                            ].append(line_time)
                            continue

                        for var_name, var_value in match.groupdict().items():
                            self.line_trackers[tracker_name]["var_data"][var_name][
                                "times"
                            ].append(line_time)
                            self.line_trackers[tracker_name]["var_data"][var_name][
                                "values"
                            ].append(float(var_value))
                        continue

            if self.time_format == TimeFormat.RELATIVE:
                self.convert_times_to_relative()

    def convert_times_to_relative(self):
        """Convert time variables to time from the first time in the log"""
        for tracker_name, tracker_data in self.line_trackers.items():
            for var_name in tracker_data["var_data"]:
                self.line_trackers[tracker_name]["var_data"][var_name]["times"] = [
                    (t - self.time_origin).seconds / 60
                    for t in self.line_trackers[tracker_name]["var_data"][var_name][
                        "times"
                    ]
                ]

    def add_figure(
        self,
        fig_name: str,
        fig_title: Optional[str] = None,
        x_label: Optional[str] = None,
        y_label: Optional[str] = None,
    ) -> None:
        """Add a new figure to plot"""
        self.figures.append(plt.figure(fig_name))
        plt.figure(fig_name)
        file_name = self.log_file_path.split("/")[-2]
        plt.title(
            f"{fig_title} [{file_name}] [Agent {self.agent_index}]"
            if fig_title
            else f"{fig_name} [{file_name}] [Agent {self.agent_index}]"
        )

        x_label = x_label if x_label else self.default_x_label
        plt.xlabel(x_label)

        if y_label:
            plt.ylabel(y_label)

    def plot(self) -> None:
        """Plot a figure"""

        # Max y values (needed for events)
        max_values: Dict[str, float] = {}
        for tracker_data in [
            v for v in self.line_trackers.values() if v["type"] != "event"
        ]:
            for figure_name in tracker_data["figure_names"]:
                for var_data in tracker_data["var_data"].values():
                    if var_data["values"]:
                        max_var = max(var_data["values"])
                        if figure_name not in max_values:
                            max_values[figure_name] = max_var
                        else:
                            if max_var > max_values[figure_name]:
                                max_values[figure_name] = max_var

        # Plot
        for tracker_name, tracker_data in self.line_trackers.items():
            for figure_name in tracker_data["figure_names"]:
                plt.figure(figure_name)
                for var_name, var_data in tracker_data["var_data"].items():
                    if tracker_data["type"] == "event" and var_data["times"]:
                        t0 = var_data["times"][0]
                        plt.plot(
                            [t0, t0],
                            [0, 1.05 * max_values[figure_name]],
                            var_data["line_style"],
                            linewidth=var_data["line_width"],
                            label=f"{tracker_name}",
                        )
                        for t in var_data["times"][1:]:
                            plt.plot(
                                [t, t],
                                [0, 1.05 * max_values[figure_name]],
                                var_data["line_style"],
                                linewidth=var_data["line_width"],
                            )
                        plt.legend()
                        continue

                    if not var_data["values"]:
                        print(
                            f"[Figure {figure_name}] Data for {tracker_name}::{var_name} not found in the log!"
                        )
                        continue

                    if not var_data["times"] or None in var_data["times"]:
                        plt.plot(
                            var_data["values"],
                            var_data["line"],
                            linewidth=var_data["line_width"],
                            label=f"{tracker_name}::{var_name}",
                        )
                    else:
                        plt.plot(
                            var_data["times"],
                            var_data["values"],
                            var_data["line_style"],
                            linewidth=var_data["line_width"],
                            label=f"{tracker_name}::{var_name}",
                        )
                    plt.legend()
        plt.show()


def extract_gc_objects_set(log_file_path):
    """Extract a set of all objects that appeared in the garbage collector count section"""
    objs = set()
    with open(log_file_path, "r") as log:
        for line in log.readlines():
            match = re.match(r".*\* (?P<name>.*) \(gc\):  (?P<value>\d+).*", line)
            if match:
                objs.add(match.groupdict()["name"])
    return sorted(objs)


def add_count_trackers(
    parser: LogParser,
    tracker_names: List[str],
    figure_names: List[str],
    tracker_name_appendix: str = "",
):
    """Basic trackers for counters"""
    for tracker_name in tracker_names:
        parser.add_tracker(
            tracker_name=f"{tracker_name}{tracker_name_appendix}",
            regex=r".*\* "
            + re.escape(f"{tracker_name}{tracker_name_appendix}")
            + r":  (?P<count>\d+).*",
            figure_names=figure_names,
        )


def main():
    """Main function"""
    log_parser = LogParser(
        agent_index=0, time_format=TimeFormat.RELATIVE
    )  # uses the most recent log by default

    # Add figures
    log_parser.add_figure(fig_name="Memory", y_label="Memory [MB]")
    log_parser.add_figure(fig_name="Object count (present)", y_label="Count")
    log_parser.add_figure(fig_name="Object count (created)", y_label="Count")
    log_parser.add_figure(fig_name="Object count (gc)", y_label="Count")

    # Memory
    log_parser.add_tracker(
        tracker_name="Memory",
        regex=r"Memory: (?P<current_memory>\d+\.\d+).*Peak (?P<peak_memory>\d+\.\d+).*",
        figure_names=["Memory"],
        tracker_type="line",
        line_styles={
            "current_memory": {
                "style": LineStyle.SOLID.value,
                "color": LineColor.GREEN.value,
                "width": 2,
            },
            "peak_memory": {
                "style": LineStyle.SOLID.value,
                "color": LineColor.RED.value,
                "width": 2,
            },
        },
    )

    # Settlement events
    log_parser.add_tracker(
        tracker_name="Settlement",
        regex=r".*Finalized with transaction hash: .*",  # validation message
        figure_names=[
            "Memory",
            "Object count (present)",
            "Object count (created)",
            "Object count (gc)",
        ],
        tracker_type="event",
        line_styles={
            "event": {
                "style": LineStyle.SOLID.value,
                "color": LineColor.BLACK.value,
                "width": 1.5,
            },
        },
    )

    # Tendermint reset events
    log_parser.add_tracker(
        tracker_name="Tendermint reset",
        regex=r".*Resetting tendermint node successful!.*",
        figure_names=[
            "Memory",
            "Object count (present)",
            "Object count (created)",
            "Object count (gc)",
        ],
        tracker_type="event",
        line_styles={
            "event": {
                "style": LineStyle.DASHED.value,
                "color": LineColor.RED.value,
                "width": 1.5,
            },
        },
    )

    # Tracked objects
    tracked_objects = [
        "Message",
        "Dialogue",
        "DialogueLabel",
        "Handler",
        "Model",
        "Behaviour",
        "Skill",
        "Connection",
        "Contract",
        "Protocol",
    ]

    # Present tracked objects
    add_count_trackers(
        parser=log_parser,
        tracker_names=tracked_objects,
        figure_names=["Object count (present)"],
        tracker_name_appendix=" (present)",
    )
    # Created tracked objects
    add_count_trackers(
        parser=log_parser,
        tracker_names=tracked_objects,
        figure_names=["Object count (created)"],
        tracker_name_appendix=" (created)",
    )
    # Present common objects in the garbage collector
    tracked_objects_in_gc = extract_gc_objects_set(log_parser.log_file_path)

    add_count_trackers(
        parser=log_parser,
        tracker_names=tracked_objects_in_gc,
        figure_names=["Object count (gc)"],
        tracker_name_appendix=" (gc)",
    )

    # Process and plot
    log_parser.process()
    log_parser.plot()


if __name__ == "__main__":
    main()
